import torch
import numpy as np
import torch.nn as nn
import torch.nn.functional as F
from torch_sparse import SparseTensor, matmul
from torch_geometric.utils import degree
# from utils import SparseLinear
from torch import Tensor
from torch.nn.parameter import Parameter
from torch.nn import init
import math

BIG_CONSTANT = 1e8
def softmax_kernel_transformation(data, is_query, projection_matrix=None, numerical_stabilizer=0.000001):
    data_normalizer = 1.0 / torch.sqrt(torch.sqrt(torch.tensor(data.shape[-1], dtype=torch.float32)))
    data = data_normalizer * data
    ratio = 1.0 / torch.sqrt(torch.tensor(projection_matrix.shape[0], dtype=torch.float32))
    data_dash = torch.einsum("bnhd,md->bnhm", data, projection_matrix)
    diag_data = torch.square(data)
    diag_data = torch.sum(diag_data, dim=len(data.shape)-1)
    diag_data = diag_data / 2.0
    diag_data = torch.unsqueeze(diag_data, dim=len(data.shape)-1)
    last_dims_t = len(data_dash.shape) - 1
    attention_dims_t = len(data_dash.shape) - 3
    if is_query:
        data_dash = ratio * (
            torch.exp(data_dash - diag_data - torch.max(data_dash, dim=last_dims_t, keepdim=True)[0]) + numerical_stabilizer
        )
    else:
        data_dash = ratio * (
            torch.exp(data_dash - diag_data - torch.max(torch.max(data_dash, dim=last_dims_t, keepdim=True)[0],
                    dim=attention_dims_t, keepdim=True)[0]) + numerical_stabilizer
        )
    return data_dash

class HyperGT(nn.Module):
    """ adapted from https://github.com/zeroxleo/HyperGT?tab=readme-ov-file """
    def __init__(self, in_channels, out_channels, args,
                 kernel_transformation=softmax_kernel_transformation):
        super().__init__()
        self.args = args

        # stem
        self.fcs = nn.ModuleList([nn.Linear(in_channels, args.hidden_channels)])
        self.bns = nn.ModuleList([nn.LayerNorm(args.hidden_channels)])
        self.convs = nn.ModuleList()

        for _ in range(args.All_num_layers):
            self.convs.append(
                NodeFormerConv(
                    args.hidden_channels, args.hidden_channels,
                    num_heads=args.num_heads,
                    kernel_transformation=kernel_transformation,
                    nb_random_features=args.nb_random_features,
                    use_gumbel=args.use_gumbel,
                    nb_gumbel_sample=args.nb_gumbel_sample,
                    rb_order=args.rb_order,
                    rb_trans=args.rb_trans,
                    use_edge_loss=args.use_edge_loss
                )
            )
            self.bns.append(nn.LayerNorm(args.hidden_channels))

        if args.use_jk:
            self.fcs.append(nn.Linear(args.hidden_channels * args.All_num_layers + args.hidden_channels, out_channels))
        else:
            self.fcs.append(nn.Linear(args.hidden_channels, out_channels))

        self.dropout = args.dropout
        self.activation = F.elu
        self.use_bn = args.use_bn
        self.use_residual = args.use_residual
        self.use_act = args.use_act
        self.use_jk = args.use_jk
        self.use_edge_loss = args.use_edge_loss
        # lazy init
        self.he_sparse_encoder = None  # in_features = m 
        self.hte_sparse_encoder = None # in_features = n

    def _ensure_pe_modules(self, H: torch.Tensor):
        """
        init he_sparse_encoder and hte_sparse_encoder
        """
        assert H.is_sparse, "H must be a sparse tensor (COO)."
        dev = H.device
        n = H.size(0)
        m = H.size(1)
        if ('HEPE' in getattr(self.args, 'pe', '')) and (self.he_sparse_encoder is None or self.he_sparse_encoder.in_features != m):
            self.he_sparse_encoder = SparseLinear(m, self.args.hidden_channels).to(dev)
        if ('HtEPE' in getattr(self.args, 'pe', '')) and (self.hte_sparse_encoder is None or self.hte_sparse_encoder.in_features != n):
            self.hte_sparse_encoder = SparseLinear(n, self.args.hidden_channels).to(dev)

    def reset_parameters(self):
        for m in self.fcs:
            if hasattr(m, 'reset_parameters'): m.reset_parameters()
        for bn in self.bns:
            if hasattr(bn, 'reset_parameters'): bn.reset_parameters()
        for conv in self.convs:
            if hasattr(conv, 'reset_parameters'): conv.reset_parameters()
   
        if self.he_sparse_encoder is not None and hasattr(self.he_sparse_encoder, 'reset_parameters'):
            self.he_sparse_encoder.reset_parameters()
        if self.hte_sparse_encoder is not None and hasattr(self.hte_sparse_encoder, 'reset_parameters'):
            self.hte_sparse_encoder.reset_parameters()

    def forward(self, data, tau=1.0):
        x = data.x  # [N, Fin]
        adjs =data.adjs
        H = data.H
        # print(f"=============forward:   {getattr(self.args, 'pe', '')}")
        if ('HEPE' in getattr(self.args, 'pe', '')) or ('HtEPE' in getattr(self.args, 'pe', '')):
            self._ensure_pe_modules(H)

        x = x.unsqueeze(0)  # [1, N, Fin]
        layer_ = []
        link_loss_=[]

        z = self.fcs[0](x)
       
        if  'HEPE' in self.args.pe:
            he_pe = self.he_sparse_encoder(H).unsqueeze(0)
            padding=torch.zeros(z.shape[0],z.shape[1]-he_pe.shape[1],z.shape[2],requires_grad=False,device=z.device)
            he_pe = torch.cat((he_pe,padding),dim=1)
            z = z + he_pe
        if 'HtEPE' in self.args.pe:
            hte_pe = self.hte_sparse_encoder(H.transpose(0,1)).unsqueeze(0)
            padding=torch.zeros(z.shape[0],z.shape[1]-hte_pe.shape[1],z.shape[2],requires_grad=False,device=z.device)
            hte_pe = torch.cat((padding,hte_pe),dim=1)
            z = z + hte_pe
        
        if self.use_bn:
            z = self.bns[0](z)
        z = self.activation(z)
        z = F.dropout(z, p=self.dropout, training=self.training)
        layer_.append(z)

        for i, conv in enumerate(self.convs):
            if self.use_edge_loss:
                z, link_loss, = conv(self.args, z, adjs, tau)
                link_loss_.append(link_loss)
            else:
                z = conv(self.args, z, adjs, tau)
            if self.use_residual:
                z += layer_[i]
            if self.use_bn:
                z = self.bns[i+1](z)
            if self.use_act:
                z = self.activation(z)
            z = F.dropout(z, p=self.dropout, training=self.training)
            layer_.append(z)

        if self.use_jk: # use jk connection for each layer
            z = torch.cat(layer_, dim=-1)

        x_out = self.fcs[-1](z).squeeze(0)
        
        num_nodes = H.size(0)
        x_out = x_out[:num_nodes]
        
        if self.use_edge_loss:
            return x_out, link_loss_
        else:
            return x_out, None



def create_projection_matrix(m, d, seed=0, scaling=0, struct_mode=False):
    nb_full_blocks = int(m/d)#0
    block_list = []
    current_seed = seed
    for _ in range(nb_full_blocks):
        torch.manual_seed(current_seed)
        if struct_mode:
            q = create_products_of_givens_rotations(d, current_seed)
        else:
            unstructured_block = torch.randn((d, d))
            q, _ = torch.qr(unstructured_block)
            q = torch.t(q)
        block_list.append(q)
        current_seed += 1
    remaining_rows = m - nb_full_blocks * d
    if remaining_rows > 0:
        torch.manual_seed(current_seed)
        if struct_mode:
            q = create_products_of_givens_rotations(d, current_seed)
        else:
            unstructured_block = torch.randn((d, d))
            q, _ = torch.qr(unstructured_block)
            q = torch.t(q)
        block_list.append(q[0:remaining_rows])
    final_matrix = torch.vstack(block_list)

    current_seed += 1
    torch.manual_seed(current_seed)
    if scaling == 0:
        multiplier = torch.norm(torch.randn((m, d)), dim=1)
    elif scaling == 1:
        multiplier = torch.sqrt(torch.tensor(float(d))) * torch.ones(m)
    else:
        raise ValueError("Scaling must be one of {0, 1}. Was %s" % scaling)

    return torch.matmul(torch.diag(multiplier), final_matrix)

class SparseLinear(nn.Module):
    r"""
    """
    __constants__ = ['in_features', 'out_features']
    in_features: int
    out_features: int
    weight: Tensor

    def __init__(self, in_features: int, out_features: int, bias: bool = True,
                 device=None, dtype=None) -> None:
        factory_kwargs = {'device': device, 'dtype': dtype}
        super(SparseLinear, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.weight = Parameter(torch.empty((out_features, in_features), **factory_kwargs))
        if bias:
            self.bias = Parameter(torch.empty(out_features, **factory_kwargs))
        else:
            self.register_parameter('bias', None)
        self.reset_parameters()

    def reset_parameters(self) -> None:
        init.kaiming_uniform_(self.weight, a=math.sqrt(5))
        # init.ones_(self.weight)

        if self.bias is not None:
            fan_in, _ = init._calculate_fan_in_and_fan_out(self.weight)
            bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
            init.uniform_(self.bias, -bound, bound)

    def forward(self, input: Tensor) -> Tensor:
        # wb=torch.sparse.mm(input,self.weight.T).to_dense()+self.bias
        wb=torch.sparse.mm(input,self.weight.T)
        if self.bias is not None:
            out = wb + self.bias
        else:
            out = wb
        return out

    def extra_repr(self) -> str:
        return 'in_features={}, out_features={}, bias={}'.format(
            self.in_features, self.out_features, self.bias is not None
        )


def create_products_of_givens_rotations(dim, seed):
    nb_givens_rotations = dim * int(math.ceil(math.log(float(dim))))
    q = np.eye(dim, dim)
    np.random.seed(seed)
    for _ in range(nb_givens_rotations):
        random_angle = math.pi * np.random.uniform()
        random_indices = np.random.choice(dim, 2)
        index_i = min(random_indices[0], random_indices[1])
        index_j = max(random_indices[0], random_indices[1])
        slice_i = q[index_i]
        slice_j = q[index_j]
        new_slice_i = math.cos(random_angle) * slice_i + math.cos(random_angle) * slice_j
        new_slice_j = -math.sin(random_angle) * slice_i + math.cos(random_angle) * slice_j
        q[index_i] = new_slice_i
        q[index_j] = new_slice_j
    return torch.tensor(q, dtype=torch.float32)

def relu_kernel_transformation(data, is_query, projection_matrix=None, numerical_stabilizer=0.001):
    del is_query
    if projection_matrix is None:
        return F.relu(data) + numerical_stabilizer
    else:
        ratio = 1.0 / torch.sqrt(
            torch.tensor(projection_matrix.shape[0], torch.float32)
        )
        data_dash = ratio * torch.einsum("bnhd,md->bnhm", data, projection_matrix)
        return F.relu(data_dash) + numerical_stabilizer



def numerator(qs, ks, vs):
    kvs = torch.einsum("nbhm,nbhd->bhmd", ks, vs) # kvs refers to U_k in the paper
    return torch.einsum("nbhm,bhmd->nbhd", qs, kvs)

def denominator(qs, ks):
    all_ones = torch.ones([ks.shape[0]]).to(qs.device)
    ks_sum = torch.einsum("nbhm,n->bhm", ks, all_ones) # ks_sum refers to O_k in the paper
    return torch.einsum("nbhm,bhm->nbh", qs, ks_sum)

def numerator_gumbel(qs, ks, vs):
    kvs = torch.einsum("nbhkm,nbhd->bhkmd", ks, vs) # kvs refers to U_k in the paper
    return torch.einsum("nbhm,bhkmd->nbhkd", qs, kvs)

def denominator_gumbel(qs, ks):
    all_ones = torch.ones([ks.shape[0]]).to(qs.device)
    ks_sum = torch.einsum("nbhkm,n->bhkm", ks, all_ones) # ks_sum refers to O_k in the paper
    return torch.einsum("nbhm,bhkm->nbhk", qs, ks_sum)

def kernelized_softmax(query, key, value, kernel_transformation, projection_matrix=None, edge_index=None, tau=0.25, return_weight=True):
    '''
    fast computation of all-pair attentive aggregation with linear complexity
    input: query/key/value [B, N, H, D]
    return: updated node emb, attention weight (for computing edge loss)
    B = graph number (always equal to 1 in Node Classification), N = node number, H = head number,
    M = random feature dimension, D = hidden size
    '''
    query = query / math.sqrt(tau)
    key = key / math.sqrt(tau)
    query_prime = kernel_transformation(query, True, projection_matrix) # [B, N, H, M]
    key_prime = kernel_transformation(key, False, projection_matrix) # [B, N, H, M]
    query_prime = query_prime.permute(1, 0, 2, 3) # [N, B, H, M]
    key_prime = key_prime.permute(1, 0, 2, 3) # [N, B, H, M]
    value = value.permute(1, 0, 2, 3) # [N, B, H, D]

    # compute updated node emb, this step requires O(N)
    z_num = numerator(query_prime, key_prime, value)
    z_den = denominator(query_prime, key_prime)

    z_num = z_num.permute(1, 0, 2, 3)  # [B, N, H, D]
    z_den = z_den.permute(1, 0, 2)
    z_den = torch.unsqueeze(z_den, len(z_den.shape))
    z_output = z_num / z_den # [B, N, H, D]

    if return_weight: # query edge prob for computing edge-level reg loss, this step requires O(E)
        start, end = edge_index
        query_end, key_start = query_prime[end], key_prime[start] # [E, B, H, M]
        edge_attn_num = torch.einsum("ebhm,ebhm->ebh", query_end, key_start) # [E, B, H]
        edge_attn_num = edge_attn_num.permute(1, 0, 2) # [B, E, H]
        attn_normalizer = denominator(query_prime, key_prime) # [N, B, H]
        edge_attn_dem = attn_normalizer[end]  # [E, B, H]
        edge_attn_dem = edge_attn_dem.permute(1, 0, 2) # [B, E, H]
        A_weight = edge_attn_num / edge_attn_dem # [B, E, H]

        return z_output, A_weight

    else:
        return z_output

def kernelized_gumbel_softmax(query, key, value, kernel_transformation, projection_matrix=None, edge_index=None,
                                K=10, tau=0.25, return_weight=False):
    '''
    fast computation of all-pair attentive aggregation with linear complexity
    input: query/key/value [B, N, H, D]
    return: updated node emb, attention weight (for computing edge loss)
    B = graph number (always equal to 1 in Node Classification), N = node number, H = head number,
    M = random feature dimension, D = hidden size, K = number of Gumbel sampling
    '''
    query = query / math.sqrt(tau)
    key = key / math.sqrt(tau)
    query_prime = kernel_transformation(query, True, projection_matrix) # [B, N, H, M]
    key_prime = kernel_transformation(key, False, projection_matrix) # [B, N, H, M]
    query_prime = query_prime.permute(1, 0, 2, 3) # [N, B, H, M]
    key_prime = key_prime.permute(1, 0, 2, 3) # [N, B, H, M]
    value = value.permute(1, 0, 2, 3) # [N, B, H, D]

    # compute updated node emb, this step requires O(N)
    gumbels = (
        -torch.empty(key_prime.shape[:-1]+(K, ), memory_format=torch.legacy_contiguous_format).exponential_().log()
    ).to(query.device) / tau # [N, B, H, K]
    key_t_gumbel = key_prime.unsqueeze(3) * gumbels.exp().unsqueeze(4) # [N, B, H, K, M]
    z_num = numerator_gumbel(query_prime, key_t_gumbel, value) # [N, B, H, K, D]
    z_den = denominator_gumbel(query_prime, key_t_gumbel) # [N, B, H, K]

    z_num = z_num.permute(1, 0, 2, 3, 4) # [B, N, H, K, D]
    z_den = z_den.permute(1, 0, 2, 3) # [B, N, H, K]
    z_den = torch.unsqueeze(z_den, len(z_den.shape))
    z_output = torch.mean(z_num / z_den, dim=3) # [B, N, H, D]

    if return_weight: # query edge prob for computing edge-level reg loss, this step requires O(E)
        start, end = edge_index
        query_end, key_start = query_prime[end], key_prime[start] # [E, B, H, M]
        edge_attn_num = torch.einsum("ebhm,ebhm->ebh", query_end, key_start) # [E, B, H]
        edge_attn_num = edge_attn_num.permute(1, 0, 2) # [B, E, H]
        attn_normalizer = denominator(query_prime, key_prime) # [N, B, H]
        edge_attn_dem = attn_normalizer[end]  # [E, B, H]
        edge_attn_dem = edge_attn_dem.permute(1, 0, 2) # [B, E, H]
        A_weight = edge_attn_num / edge_attn_dem # [B, E, H]

        return z_output, A_weight

    else:
        return z_output

def add_conv_relational_bias(x, edge_index, b, trans='sigmoid'):
    '''
    compute updated result by the relational bias of input adjacency
    the implementation is similar to the Graph Convolution Network with a (shared) scalar weight for each edge
    '''
    row, col = edge_index
    d_in = degree(col, x.shape[1]).float()
    d_norm_in = (1. / d_in[col]).sqrt()
    d_out = degree(row, x.shape[1]).float()
    d_norm_out = (1. / d_out[row]).sqrt()
    conv_output = []
    for i in range(x.shape[2]):
        if trans == 'sigmoid':
            b_i = F.sigmoid(b[i])
        elif trans == 'identity':
            b_i = b[i]
        else:
            raise NotImplementedError
        value = torch.ones_like(row) * b_i * d_norm_in * d_norm_out
        adj_i = SparseTensor(row=col, col=row, value=value, sparse_sizes=(x.shape[1], x.shape[1]))
        conv_output.append( matmul(adj_i, x[:, :, i]) )  # [B, N, D]
    conv_output = torch.stack(conv_output, dim=2) # [B, N, H, D]
    return conv_output

class NodeFormerConv(nn.Module):
    '''
    one layer of NodeFormer that attentive aggregates all nodes over a latent graph
    return: node embeddings for next layer, edge loss at this layer
    '''
    def __init__(self, in_channels, out_channels, num_heads, kernel_transformation=softmax_kernel_transformation, projection_matrix_type='a',
                 nb_random_features=10, use_gumbel=True, nb_gumbel_sample=10, rb_order=0, rb_trans='sigmoid', use_edge_loss=True):
        super(NodeFormerConv, self).__init__()
        self.Wk = nn.Linear(in_channels, out_channels * num_heads)
        self.Wq = nn.Linear(in_channels, out_channels * num_heads)
        self.Wv = nn.Linear(in_channels, out_channels * num_heads)
        self.Wo = nn.Linear(out_channels * num_heads, out_channels)
        if rb_order >= 1:
            self.b = torch.nn.Parameter(torch.FloatTensor(rb_order, num_heads), requires_grad=True)

        self.out_channels = out_channels
        self.num_heads = num_heads
        self.kernel_transformation = kernel_transformation
        self.projection_matrix_type = projection_matrix_type
        self.nb_random_features = nb_random_features
        self.use_gumbel = use_gumbel
        self.nb_gumbel_sample = nb_gumbel_sample
        self.rb_order = rb_order
        self.rb_trans = rb_trans
        self.use_edge_loss = use_edge_loss

    def reset_parameters(self):
        self.Wk.reset_parameters()
        self.Wq.reset_parameters()
        self.Wv.reset_parameters()
        self.Wo.reset_parameters()
        if self.rb_order >= 1:
            if self.rb_trans == 'sigmoid':
                torch.nn.init.constant_(self.b, 0.1)
            elif self.rb_trans == 'identity':
                torch.nn.init.constant_(self.b, 1.0)

    def forward(self, args, z, adjs, tau):
        B, N = z.size(0), z.size(1)
        query = self.Wq(z).reshape(-1, N, self.num_heads, self.out_channels)
        key = self.Wk(z).reshape(-1, N, self.num_heads, self.out_channels)
        value = self.Wv(z).reshape(-1, N, self.num_heads, self.out_channels)

        if self.projection_matrix_type is None:
            projection_matrix = None
        else:
            dim = query.shape[-1]
            seed = torch.ceil(torch.abs(torch.sum(query) * BIG_CONSTANT)).to(torch.int32)
            # seed = 0
            projection_matrix = create_projection_matrix(
                self.nb_random_features, dim, seed=seed).to(query.device)

        # compute all-pair message passing update and attn weight on input edges, requires O(N) or O(N + E)

        if self.use_gumbel and self.training:  # only using Gumbel noise for training
            if self.use_edge_loss:
                z_next, weight = kernelized_gumbel_softmax(query,key,value,self.kernel_transformation,projection_matrix,adjs[0],
                                                    self.nb_gumbel_sample, tau, self.use_edge_loss)
            else:
                z_next = kernelized_gumbel_softmax(query,key,value,self.kernel_transformation,projection_matrix,adjs[0],
                                                    self.nb_gumbel_sample, tau, self.use_edge_loss)
        else:
            if self.use_edge_loss:
                z_next, weight = kernelized_softmax(query, key, value, self.kernel_transformation, projection_matrix, adjs[0],
                                                    tau, self.use_edge_loss)
            else:
                z_next = kernelized_gumbel_softmax(query,key,value,self.kernel_transformation,projection_matrix,adjs[0],
                                                    self.nb_gumbel_sample, tau, self.use_edge_loss)
       
        # compute update by relational bias of input adjacency, requires O(E)
        for i in range(self.rb_order):
            z_next += add_conv_relational_bias(value, adjs[i], self.b[i], self.rb_trans)

        # aggregate results of multiple heads
        z_next = self.Wo(z_next.flatten(-2, -1))

        if self.use_edge_loss: # compute edge regularization loss on input adjacency
            row, col = adjs[0]
            d_in = degree(col, query.shape[1]).float()
            d_norm = 1. / d_in[col]
            d_norm_ = d_norm.reshape(1, -1, 1).repeat(1, 1, weight.shape[-1])
            link_loss = torch.mean(weight.log() * d_norm_)
            return z_next, link_loss
        else:
            return z_next




import torch.nn as nn 
import torch.nn.functional as F
from torch_geometric.nn import GCNConv,GATConv

class CEGCN(nn.Module):
    def __init__(self,num_features, num_targets, args):
        super(CEGCN, self).__init__()
        
        self.hidden_dim = args.MLP_hidden
        self.num_layers = args.All_num_layers
        self.dropout=args.dropout
        self.use_bn=args.use_bn

        self.convs = nn.ModuleList([])
        self.bns = nn.ModuleList()

        if self.num_layers == 1:
            self.convs.append(GCNConv(in_channels=num_features, out_channels=num_targets))
        else:
            self.convs.append(GCNConv(in_channels=num_features, out_channels=self.hidden_dim))
            if self.use_bn:
                self.bns.append(nn.BatchNorm1d(self.hidden_dim))
            for l in range(self.num_layers - 2):
                self.convs.append(GCNConv(in_channels=self.hidden_dim, out_channels=self.hidden_dim))
                if self.use_bn:
                    self.bns.append(nn.BatchNorm1d(self.hidden_dim))
            self.convs.append(GCNConv(in_channels=self.hidden_dim, out_channels=num_targets))

    def reset_parameters(self):
        for layer in self.convs:
            layer.reset_parameters()

        if self.use_bn:
            for bn in self.bns:
                bn.reset_parameters()

    def forward(self, data):
        x, edge_index = data.x, data.clique_edge_index
        for i, conv in enumerate(self.convs[:-1]):
            x = conv(x, edge_index)
            if self.use_bn:
                x = self.bns[i](x)
            x = F.relu(x, inplace=True)
            x = F.dropout(x, p=self.dropout, training=self.training)
        x = self.convs[-1](x, edge_index)
        return x, None


class CEGAT(nn.Module):
    def __init__(self,num_features, num_targets, args):
        super(CEGAT, self).__init__()
        
        self.hidden_dim = args.MLP_hidden
        self.num_layers = args.All_num_layers
        self.dropout=args.dropout

        self.heads = args.heads
        self.use_bn = args.use_bn
        self.concat = args.concat

        self.convs = nn.ModuleList([])
        self.bns = nn.ModuleList()

        if self.num_layers == 1:
            self.convs.append(GATConv(in_channels=num_features, 
                                      out_channels=num_targets,
                                      heads=1,
                                      concat=False)) 
        else:
            self.convs.append(GATConv(in_channels=num_features, 
                                      out_channels=self.hidden_dim,
                                      heads=self.heads,
                                      concat=self.concat))
            if self.use_bn:
                dim1 = self.hidden_dim * self.heads if self.concat else self.hidden_dim
                self.bns.append(nn.BatchNorm1d(dim1))
                        
            for l in range(self.num_layers - 2):
                in_dim = self.hidden_dim * self.heads if self.concat else self.hidden_dim
                self.convs.append(GATConv(in_channels=in_dim, 
                                          out_channels=self.hidden_dim,
                                          heads=self.heads,
                                          concat=self.concat))
                if self.use_bn:
                    out_dim = self.hidden_dim * self.heads if self.concat else self.hidden_dim
                    self.bns.append(nn.BatchNorm1d(out_dim))
            
            final_in = self.hidden_dim * self.heads if self.concat else self.hidden_dim
            self.convs.append(GATConv(in_channels=final_in, 
                                      out_channels=num_targets,
                                      heads=1,
                                      concat=False))

    def reset_parameters(self):
        for layer in self.convs:
            layer.reset_parameters()
        if self.use_bn:
            for bn in self.bns:
                bn.reset_parameters()

    def forward(self, data):
        x, edge_index = data.x, data.clique_edge_index
        for i, conv in enumerate(self.convs[:-1]):
            x = conv(x, edge_index)
            if self.use_bn:
                x = self.bns[i](x)
            x = F.relu(x, inplace=True)
            x = F.dropout(x, p=self.dropout, training=self.training)
        x = self.convs[-1](x, edge_index)
        return x, None


import torch
import torch.nn as nn
from dhgbench.lib_models.HNN.mlp import MLP
import torch_scatter

class HyperND(nn.Module):
    def __init__(self, num_features, num_targets, args):

        super().__init__()

        self.hidden_dim = args.MLP_hidden
        self.num_layers = args.MLP_num_layers 
        self.dropout=args.dropout
        self.norm = args.normalization

        self.p = args.HyperND_ord
        self.tol = args.HyperND_tol
        self.max_steps = args.HyperND_steps
        self.aggr = args.aggregate
        self.alpha = args.restart_alpha

        self.cache = None

        self.decoder = MLP(in_channels=num_features,
            hidden_channels=self.hidden_dim,
            out_channels=num_targets,
            num_layers=self.num_layers,
            dropout=self.dropout,
            Normalization=self.norm,
            InputNorm=False)

    def reset_parameters(self):
        self.cache = None
        self.decoder.reset_parameters()

    def diffusion(self, data):
        if self.cache is not None:
            return self.cache

        X = data.x
        # Y = nn.functional.one_hot(data.y)
        N = X.shape[-2]
        V, E = data.hyperedge_index[0], data.hyperedge_index[1]
        device = X.device

        ones = torch.ones(V.shape[0], device=device, dtype=torch.float32)
        D = torch_scatter.scatter_add(ones, V, dim=0)

        ones = torch.ones(E.shape[0], device=device, dtype=torch.float32)
        DE = torch_scatter.scatter_add(ones, E, dim=0)

        def rho(Z):
            return torch.pow(Z, self.p)

        def sigma(Z):
            Z = Z / DE[:, None]
            return torch.pow(Z, 1. / self.p)

        def V2E(Z):
            Z = Z / torch.pow(D, 0.5)[:, None]
            Xve = rho(Z[..., V, :]) # [nnz, C] 
            Xe = torch_scatter.scatter(Xve, E, dim=-2, reduce=self.aggr) # [E, C]
            Xe = sigma(Xe)
            return Xe

        def E2V(Z):
            Xev = Z[..., E, :] # [nnz, C]
            Xv = torch_scatter.scatter(Xev, V, dim=-2, reduce=self.aggr, dim_size=N) # [N, C]
            Xv = Xv / torch.pow(D, 0.5)[:, None]
            return Xv

        def phi(Z):
            P = V2E(Z)
            P = torch.linalg.norm(P, dim=-1, ord=2.) ** 2.
            return 2. * torch.sqrt(P.sum())

        # U = torch.cat([X, Y], -1) + 1e-6
        U = X / phi(X)
        F = U
        for i in range(self.max_steps):
            F_last = F

            G = (1. - self.alpha) * E2V(V2E(F)) + self.alpha * U
            F = G / phi(G)

            d = torch.linalg.norm(F - F_last, ord='fro') / torch.linalg.norm(F, ord='fro')
            if d < self.tol:
                print(f'Interrupt hypergraph diffusion with {i} iterations and d={d}')
                break

        self.cache = F

        return F

    def forward(self, data):
        F = self.diffusion(data)
        return self.decoder(F), None 


from tqdm import tqdm
import torch
from collections import defaultdict
import numpy as np
import math

def neg_generator(HE, pred_num):
    # mns = MNSSampler(pred_num)
    mns = MNSSampler(pred_num)
    sns = SNSSampler(pred_num)
    cns = CNSSampler(pred_num)
    
    t_mns = mns(set(HE))
    t_sns = sns(set(HE))
    t_cns = cns(set(HE))
    
    t_mns = list(t_mns)
    t_sns = list(t_sns)
    t_cns = list(t_cns)
    
    t_mns = [list(edge) for edge in t_mns]
    t_sns = [list(edge) for edge in t_sns]
    t_cns = [list(edge) for edge in t_cns]
    
    return t_mns, t_sns, t_cns

def negative_sample(
        nodes_to_neighbors, size_dist, num_negative, hyperedges, method, corrupt_num = 1, half=False, rw_path = None):
    nodes = list(nodes_to_neighbors.keys())
    neg_samples = []
    if method == 'UNS':
        size_dist = get_pure_sample_size_dist(len(nodes))
    
    size_dist = size_dist.items()  
    if method in ['SNS', 'UNS']:
        for i in tqdm(range(num_negative), leave=False):
            sampled_edge = sized_random_sampling(
                size_dist, nodes, nodes_to_neighbors, hyperedges)
            neg_samples.append(sampled_edge)
    elif method == 'MNS':
        for i in tqdm(range(num_negative), leave=False):
            sampled_node = sized_mf_sampling(
                size_dist, nodes, nodes_to_neighbors, hyperedges)
            neg_samples.append(sampled_node)     
    elif method == 'CNS':
        list_hyperedges = list(hyperedges)
        node_set = set(nodes_to_neighbors.keys())
        for i in tqdm(range(num_negative), leave=False):
            sampled_edge = clique_negative_sampling(
                hyperedges, nodes_to_neighbors, num_negative, list_hyperedges,
                node_set)
            neg_samples.append(sampled_edge)       

    return neg_samples

def generate_negative_samples_for_hyperedges(
        hyperedges, method, neg_samples_size, corrupt_num = 1, half=False, rw_path = None):
    #print(hyperedges)
    edges = {
        frozenset({u, v}) for hedge in hyperedges
        for u in hedge for v in hedge if u > v}
    nodes_to_neighbors = defaultdict(set)
    for edge in edges:
        u, v = edge
        nodes_to_neighbors[u].add(v)
        nodes_to_neighbors[v].add(u)

    size_dist = generate_hyperedge_size_dist(hyperedges)
    
    #print('Generating Negative Samples')
    total = math.ceil(neg_samples_size)
    neg_samples = negative_sample(
        nodes_to_neighbors, size_dist, total, hyperedges, method, corrupt_num = 1, half=False, rw_path = rw_path)
    negative_hyperedges = [frozenset(x) for x, y in neg_samples]
    
    return negative_hyperedges

class SNSSampler(object):
    def __init__(self, pred_num):
        self.pred_num = pred_num
    def __call__(self, hedges):
        neg_samples_size = int(self.pred_num)
        neg_samples = generate_negative_samples_for_hyperedges(hedges, 'SNS', neg_samples_size)
        return neg_samples

class MNSSampler(object):
    def __init__(self, pred_num):
        self.pred_num = pred_num
    def __call__(self, hedges):
        neg_samples_size = int(self.pred_num)
        neg_samples = generate_negative_samples_for_hyperedges(hedges, 'MNS', neg_samples_size)
        return neg_samples
    
class CNSSampler(object):
    def __init__(self, pred_num):
        self.pred_num = pred_num
    def __call__(self, hedges):
        neg_samples_size = int(self.pred_num)
        neg_samples = generate_negative_samples_for_hyperedges(hedges, 'CNS', neg_samples_size)
        return neg_samples

'''-------------SNS Sampling Utils------------------'''

def get_pure_sample_size_dist(num_nodes):
    N = num_nodes
    nck = 1
    size_dist = dict()
    size_dist[0] = 1
    for idx in range(1, num_nodes):
        nck *= (N - (idx - 1)) / idx
        size_dist[idx] = nck
    size_dist[N] = 1
    total = sum(v for k, v in size_dist.items())
    for i in size_dist:
        size_dist[i] = float(size_dist[i]) / total
    return size_dist

def sized_random_sampling(size_dist, nodes, nodes_to_neighbors, hyperedges):
    vals = [v for v, p in size_dist]
    p = [p for v, p in size_dist]
    sampled_nodes = frozenset({'a', 'b'})
    hyperedges.add(sampled_nodes)
    sampled_size = np.random.choice(vals, p=p)
    while frozenset(sampled_nodes) in hyperedges:
        sampled_nodes = [
            nodes[idx] for idx in np.random.choice(
                len(nodes), size=sampled_size, replace=False)]
    edges = {
        frozenset([node, node2])
        for node in sampled_nodes for node2 in sampled_nodes
        if node2 in nodes_to_neighbors[node] and node2 < node}
    hyperedges.remove(frozenset({'a', 'b'}))
    return sampled_nodes, edges

def generate_hyperedge_size_dist(hyperedges):
    size_dist = defaultdict(int)
    for edge in hyperedges:
        size_dist[len(edge)] += 1
    if 1 in size_dist:
        del size_dist[1]
    if 2 in size_dist:
        del size_dist[2]
    total = sum(v for k, v in size_dist.items())
    for i in size_dist:
        size_dist[i] = float(size_dist[i]) / total
    return size_dist

'''-------------MNS Sampling Utils------------------'''

def sample_initial_edge(nodes_to_neighbors):
    edgeidx = np.random.choice(
        sum(len(nodes_to_neighbors[n]) for n in nodes_to_neighbors))
    carry = 0
    for n in nodes_to_neighbors:
        if edgeidx < carry + len(nodes_to_neighbors[n]):
            edge = [n, list(nodes_to_neighbors[n])[edgeidx - carry]]
            break
        carry += len(nodes_to_neighbors[n])
    return edge

def mfinder_sampling(nodes_to_neighbors, k):
    neighbor_edges = []
    induced_edges = set()
    sampled_nodes = set()
    
    while len(sampled_nodes) < k:
        while len(neighbor_edges) == 0:
            edge = sample_initial_edge(nodes_to_neighbors)
            sampled_nodes = set(edge)
            neighbor_edges = set([
                frozenset([node, nnode])
                for node in sampled_nodes for nnode in nodes_to_neighbors[node]
                if nnode not in sampled_nodes])
            neighbor_edges = list(neighbor_edges)

            induced_edges = set()
            induced_edges.add(frozenset(edge))
        
        selected_edge = neighbor_edges[np.random.choice(len(neighbor_edges))]
        induced_edges.add(selected_edge)
        new_node = [n for n in selected_edge.difference(sampled_nodes)][0]
        sampled_nodes.add(new_node)
        
        neighbor_edges = [
            edge for edge in neighbor_edges if new_node not in edge]
        
        new_edges = set()
        
        for node in nodes_to_neighbors[new_node]:
            if node not in sampled_nodes:
                new_edges.add(frozenset([new_node, node]))
            else:
                induced_edges.add(frozenset([new_node, node]))
        
        neighbor_edges.extend(list(new_edges))    
        #assert len(neighbor_edges) == len(set(neighbor_edges))

    return sampled_nodes, induced_edges

def sized_mf_sampling(size_dist, nodes, nodes_to_neighbors, hyperedges):
    vals = [v for v, p in size_dist]
    p = [p for v, p in size_dist]
    sampled_size = np.random.choice(vals, p=p)
    sampled_nodes = {'a', 'b'}
    hyperedges.add(frozenset(sampled_nodes))
    '''while frozenset(sampled_nodes) in hyperedges:
        sampled_nodes, sampled_edge = mfinder_sampling(
            nodes_to_neighbors, sampled_size)'''
    sampled_nodes= mfinder_sampling(nodes_to_neighbors, sampled_size)
    hyperedges.remove(frozenset({'a', 'b'}))
    return sampled_nodes

'''-------------CNS Sampling Utils------------------'''

def clique_negative_sampling(
        hyperedges, nodes_to_neighbors, num_negative,
        list_hyperedges, node_set):
    edgeidx = np.random.choice(len(hyperedges), size=1)[0]
    neg = list_hyperedges[edgeidx]

    while neg in hyperedges:
        edgeidx = np.random.choice(len(hyperedges), size=1)[0]
        edge = list(list_hyperedges[edgeidx])
        node_to_remove = np.random.choice(len(edge), size=1)[0]
        nodes_to_keep = edge[:node_to_remove] + edge[node_to_remove+1:]
        probable_neighbors = node_set
        for node in nodes_to_keep:
            probable_neighbors = probable_neighbors.intersection(
                nodes_to_neighbors[node])
        
        if len(probable_neighbors) == 0:
            continue
        probable_neighbors = list(probable_neighbors)
        #check_probable_neighbors = torch.stack(probable_neighbors)
        check_probable_neighbors = torch.tensor(probable_neighbors)
        neighbor_node = np.random.choice(check_probable_neighbors.cpu(), size=1)[0]
        
        nodes_to_keep.append(neighbor_node)
        neg = frozenset(nodes_to_keep)

    edges = {
        frozenset([node1, node2])
        for node1 in neg for node2 in neg if node1 < node2
    }
    return neg, edges


import torch
from torch.nn import Linear
import torch.nn as nn
import torch.nn.functional as F
from torch_geometric.nn.conv import MessagePassing

class HNHNConv(MessagePassing):
    def __init__(self, in_channels, hidden_channels, out_channels, heads=1, nonlinear_inbetween=True,
                 concat=True, bias=True, **kwargs):
        kwargs.setdefault('aggr', 'add')
        super(HNHNConv, self).__init__(node_dim=0, **kwargs)

        self.in_channels = in_channels
        self.hidden_channels = hidden_channels
        self.out_channels = out_channels
        self.nonlinear_inbetween = nonlinear_inbetween

        self.heads = heads
        self.concat = True

        self.weight_v2e = Linear(in_channels, hidden_channels)
        self.weight_e2v = Linear(hidden_channels, out_channels)

        self.reset_parameters()

    def reset_parameters(self):
        self.weight_v2e.reset_parameters()
        self.weight_e2v.reset_parameters()

    def forward(self, x, data):

        hyperedge_index = data.hyperedge_index
        hyperedge_weight = None
        num_nodes, num_edges = x.size(0), 0
        if hyperedge_index.numel() > 0:
            num_edges = int(hyperedge_index[1].max()) + 1

        if hyperedge_weight is None:
            hyperedge_weight = x.new_ones(num_edges)

        x = self.weight_v2e(x)

        x = data.D_v_beta.unsqueeze(-1) * x

        self.flow = 'source_to_target'
        out = self.propagate(hyperedge_index, x=x, norm=data.D_e_beta_inv,
                             size=(num_nodes, num_edges))
        
        edge_embed=out 

        if self.nonlinear_inbetween:
            out = F.relu(out)
        
        # sanity check
        out = torch.squeeze(out, dim=1)
        
        out = self.weight_e2v(out)
        
        out = data.D_e_alpha.unsqueeze(-1) * out

        self.flow = 'target_to_source'
        out = self.propagate(hyperedge_index, x=out, norm=data.D_v_alpha_inv,
                             size=(num_nodes,num_edges))
        
        return out, edge_embed

    def message(self, x_j, norm_i):

        out = norm_i.view(-1, 1) * x_j

        return out

    def __repr__(self):
        return "{}({}, {}, {})".format(self.__class__.__name__, self.in_channels,
                                   self.hidden_channels, self.out_channels)

class HNHN(nn.Module):

    def __init__(self, num_features, num_targets, args):
        super(HNHN, self).__init__()

        self.num_layers = args.All_num_layers
        self.dropout = args.dropout
        self.hidden_dim = args.MLP_hidden
        
        self.convs = nn.ModuleList()
        # two cases
        if self.num_layers == 1:
            self.convs.append(HNHNConv(num_features, self.hidden_dim, num_targets,
                                       nonlinear_inbetween=args.HNHN_nonlinear_inbetween))
        else:
            self.convs.append(HNHNConv(num_features, self.hidden_dim, self.hidden_dim,
                                       nonlinear_inbetween=args.HNHN_nonlinear_inbetween))
            for _ in range(self.num_layers - 2):
                self.convs.append(HNHNConv(self.hidden_dim, self.hidden_dim, self.hidden_dim,
                                           nonlinear_inbetween=args.HNHN_nonlinear_inbetween))
            self.convs.append(HNHNConv(self.hidden_dim, self.hidden_dim, num_targets,
                                       nonlinear_inbetween=args.HNHN_nonlinear_inbetween))

    def reset_parameters(self):
        for conv in self.convs:
            conv.reset_parameters()

    def forward(self, data):

        x = data.x
        
        if self.num_layers == 1:
            conv = self.convs[0]
            x,e = conv(x, data)
        else:
            for i, conv in enumerate(self.convs[:-1]):
                x,e = conv(x, data)
                x = F.relu(x)
                x = F.dropout(x, p=self.dropout, training=self.training)
            x,e = self.convs[-1](x, data)

        return x,e


import copy
import torch
import numpy as np
from collections import Counter
from torch_scatter import scatter_add
from torch_geometric.nn.conv.gcn_conv import gcn_norm

'''-----------------Hypernode/edge-level Task Datasets Preprocessing'''

def data_processing(args,db):
    
    data=copy.deepcopy(db.data)
    
    if args.method in ['AllSetformer', 'AllDeepSets']:
        data = ExtractV2E(data)
        if args.add_self_loop:
            data = Add_Self_Loops(data)
        if args.exclude_self:
            data = expand_edge_index(data)
        data = norm_contruction(data, option=args.normtype)
        db.norm=data.norm.to(args.device)
    else:
        data = ExtractV2E(data)
        if args.add_self_loop:
            data = Add_Self_Loops(data)
        data.edge_index[1] -= data.edge_index[1].min()
    
    if db.sens is not None:
        db.sens=db.sens.to(args.device)
    
    db.x=data.x.to(args.device)
    db.hyperedge_index=data.edge_index.to(args.device)
    db.y=data.y.to(args.device)
    db.data=data
    
    return db

def expand_edge_index(data, edge_th=0):

    edge_index = data.edge_index
    num_nodes = data.n_x[0].item()
    if hasattr(data, 'totedges'):
        num_edges = data.totedges
    else:
        num_edges = data.num_hyperedges[0]

    expanded_n2he_index = []

    cur_he_id = num_nodes

    new_edge_id_2_original_edge_id = {}

    for he_idx in range(num_nodes, num_edges + num_nodes):
        # find all nodes within the same hyperedge.
        selected_he = edge_index[:, edge_index[1] == he_idx]
        size_of_he = selected_he.shape[1]

#         Trim a hyperedge if its size>edge_th
        if edge_th > 0:
            if size_of_he > edge_th:
                continue

        if size_of_he == 1:

            new_n2he = selected_he.clone()
            new_n2he[1] = cur_he_id
            expanded_n2he_index.append(new_n2he)

            cur_he_id += 1
            continue

        new_n2he = selected_he.repeat_interleave(size_of_he, dim=1)

        # new_edge_ids start from the he_id from previous iteration (cur_he_id).
        new_edge_ids = torch.LongTensor(
            np.arange(cur_he_id, cur_he_id + size_of_he)).repeat(size_of_he)
        new_n2he[1] = new_edge_ids

        # build a mapping between node and it's corresponding edge.
        # e.g. {n_1: e_7_1, n_2: e_7_2}
        tmp_node_id_2_he_id_dict = {}
        for idx in range(size_of_he):
            new_edge_id_2_original_edge_id[cur_he_id] = he_idx
            cur_node_id = selected_he[0][idx].item()
            tmp_node_id_2_he_id_dict[cur_node_id] = cur_he_id
            cur_he_id += 1

        # create n2he by deleting the self-product edge.
        new_he_select_mask = torch.BoolTensor([True] * new_n2he.shape[1])
        for col_idx in range(new_n2he.shape[1]):
            tmp_node_id, tmp_edge_id = new_n2he[0, col_idx].item(
            ), new_n2he[1, col_idx].item()
            if tmp_node_id_2_he_id_dict[tmp_node_id] == tmp_edge_id:
                new_he_select_mask[col_idx] = False
        new_n2he = new_n2he[:, new_he_select_mask]
        expanded_n2he_index.append(new_n2he)

    new_edge_index = torch.cat(expanded_n2he_index, dim=1)

    new_order = new_edge_index[0].argsort()
    data.edge_index = new_edge_index[:, new_order]

    return data

def ExtractV2E(data):
    # Assume edge_index = [V|E;E|V]
    edge_index = data.edge_index
#     First, ensure the sorting is correct (increasing along edge_index[0])
    _, sorted_idx = torch.sort(edge_index[0])
    edge_index = edge_index[:, sorted_idx].type(torch.LongTensor)

    num_nodes = data.n_x
    num_hyperedges = data.num_hyperedges
    #if not ((data.n_x[0]+data.num_hyperedges[0]-1) == data.edge_index[0].max().item()):
    if not ((data.n_x+data.num_hyperedges-1) == data.edge_index[0].max().item()):
        print('num_hyperedges does not match! 1')
        return
    cidx = torch.where(edge_index[0] == num_nodes)[
        0].min()  # cidx: [V...|cidx E...]
    data.edge_index = edge_index[:, :cidx].type(torch.LongTensor)
    return data

def Add_Self_Loops(data):
    
    # update so we dont jump on some indices
    # Assume edge_index = [V;E]. If not, use ExtractV2E()
    edge_index = data.edge_index
    num_nodes = data.n_x
    num_hyperedges = data.num_hyperedges

    if not ((data.n_x + data.num_hyperedges - 1) == data.edge_index[1].max().item()):
        print('num_hyperedges does not match! 2')
        return

    hyperedge_appear_fre = Counter(edge_index[1].numpy())
    # store the nodes that already have self-loops
    skip_node_lst = []
    for edge in hyperedge_appear_fre:
        if hyperedge_appear_fre[edge] == 1:
            skip_node = edge_index[0][torch.where(
                edge_index[1] == edge)[0].item()]
            skip_node_lst.append(skip_node.item())

    new_edge_idx = edge_index[1].max() + 1
    new_edges = torch.zeros(
        (2, num_nodes - len(set(skip_node_lst))), dtype=edge_index.dtype) 
    tmp_count = 0
    for i in range(num_nodes):
        if i not in skip_node_lst:
            new_edges[0][tmp_count] = i
            new_edges[1][tmp_count] = new_edge_idx
            new_edge_idx += 1
            tmp_count += 1

    data.totedges = num_hyperedges + num_nodes - len(set(skip_node_lst)) 
    edge_index = torch.cat((edge_index, new_edges), dim=1)
    # Sort along w.r.t. nodes
    _, sorted_idx = torch.sort(edge_index[0])
    data.edge_index = edge_index[:, sorted_idx].type(torch.LongTensor)
    return data

def norm_contruction(data, option='all_one', TYPE='V2E'):
    if TYPE == 'V2E':
        if option == 'all_one':
            data.norm = torch.ones_like(data.edge_index[0])

        elif option == 'deg_half_sym':
            edge_weight = torch.ones_like(data.edge_index[0])
            cidx = data.edge_index[1].min()
            Vdeg = scatter_add(edge_weight, data.edge_index[0], dim=0)
            HEdeg = scatter_add(edge_weight, data.edge_index[1]-cidx, dim=0)
            V_norm = Vdeg**(-1/2)
            E_norm = HEdeg**(-1/2)
            data.norm = V_norm[data.edge_index[0]] * \
                E_norm[data.edge_index[1]-cidx]

    elif TYPE == 'V2V':
        data.edge_index, data.norm = gcn_norm(
            data.edge_index, data.norm, add_self_loops=True)
    return data

def rand_train_test_idx(label, train_prop=.5, valid_prop=.25, ignore_negative=True, balance=False):
    """ Adapted from https://github.com/CUAI/Non-Homophily-Benchmarks"""
    """ randomly splits label into train/valid/test splits """
    if not balance:
        if ignore_negative:
            labeled_nodes = torch.where(label != -1)[0]
        else:
            labeled_nodes = label

        n = labeled_nodes.shape[0]
        train_num = int(n * train_prop)
        valid_num = int(n * valid_prop)

        perm = torch.as_tensor(np.random.permutation(n))

        train_indices = perm[:train_num]
        val_indices = perm[train_num:train_num + valid_num]
        test_indices = perm[train_num + valid_num:]

        if not ignore_negative:
            return train_indices, val_indices, test_indices

        train_idx = labeled_nodes[train_indices]
        valid_idx = labeled_nodes[val_indices]
        test_idx = labeled_nodes[test_indices]

        split_idx = {'train': train_idx,
                     'valid': valid_idx,
                     'test': test_idx}
    else:
        #         ipdb.set_trace()
        indices = []
        for i in range(label.max()+1):
            index = torch.where((label == i))[0].view(-1)
            index = index[torch.randperm(index.size(0))]
            indices.append(index)

        percls_trn = int(train_prop/(label.max()+1)*len(label))
        val_lb = int(valid_prop*len(label))
        train_idx = torch.cat([i[:percls_trn] for i in indices], dim=0)
        rest_index = torch.cat([i[percls_trn:] for i in indices], dim=0)
        rest_index = rest_index[torch.randperm(rest_index.size(0))]
        valid_idx = rest_index[:val_lb]
        test_idx = rest_index[val_lb:]
        split_idx = {'train': train_idx,
                     'valid': valid_idx,
                     'test': test_idx}
    return split_idx

'''-----------------Hypergraph-level Task Datasets Preprocessing'''

def convert_to_hyperedge_index(e_list):
    node_list = []
    edge_list = []
    for edge_idx, cur_he in enumerate(e_list):
        cur_size = len(cur_he)

        node_list += cur_he 
        edge_list += [edge_idx] * cur_size

    edge_index = np.array([node_list,edge_list], dtype = int)
    edge_index = torch.LongTensor(edge_index)
    return edge_index

def compute_degree_list(hyperedge_index):
    node_ids = hyperedge_index[0]  
    max_node_id = node_ids.max().item()  
    degrees = torch.bincount(node_ids, minlength=max_node_id + 1) 
    return degrees.tolist()  



import numpy as np
import torch
import torch.nn.functional as F
from torch import Tensor
from sklearn.metrics import f1_score, roc_auc_score,average_precision_score,accuracy_score
from collections import defaultdict
from dhgbench.lib_models.HNN.preprocessing import algo_preprocessing
from dhgbench.lib_utils.utils import relabel_hyperedge_index,add_self_loop_hyperedges

def eval_acc(y_true, y_pred):
    acc_list = []
    y_true = y_true.detach().cpu().numpy()
    y_pred = y_pred.argmax(dim=-1, keepdim=False).detach().cpu().numpy()

#     ipdb.set_trace()
#     for i in range(y_true.shape[1]):
    is_labeled = y_true == y_true
    correct = y_true[is_labeled] == y_pred[is_labeled]
    acc_list.append(float(np.sum(correct))/len(correct))

    return sum(acc_list)/len(acc_list)

@torch.no_grad()
def evaluate(model, data, split_idx, result=None):
    if result is not None:
        out = result
    else:
        model.eval()
        out,_ = model(data)
        out = F.log_softmax(out, dim=1)

    train_acc = eval_acc(
        data.y[split_idx['train']], out[split_idx['train']])
    valid_acc = eval_acc(
        data.y[split_idx['valid']], out[split_idx['valid']])
    test_acc = eval_acc(
        data.y[split_idx['test']], out[split_idx['test']])
    
    return train_acc, valid_acc, test_acc

def evaluate_hypegraph(model,batch_loaders,args):
    metrics_dict = defaultdict(list)
    for key,data_loaders in batch_loaders.items():
        acc,macro_f1 = test_hypegraph_loader(model,data_loaders,args)
        metrics_dict['acc'].append(acc)
        metrics_dict['macro_f1'].append(macro_f1)
    return metrics_dict 

def test_hypegraph_loader(model,data_loader,args):
    
    device = args.device
    
    model.eval()
    all_preds = []
    all_labels = []

    with torch.no_grad():

        for batch in data_loader:
            
            if args.method in ['HyperND','TFHNN']:
                model.encoder.cache=None 
            elif args.method in ['HyperGCN']:
                model.encoder.structure=None
            elif args.method in ['SheafHyperGNN']:
                model.encoder.hyperedge_attr=None 

            if args.method in ['HyperND','TFHNN','HyperGCN','SheafHyperGNN']:
                reindex_hyperedge_index,_ = relabel_hyperedge_index(batch.hyperedge_index)
                batch.hyperedge_index = reindex_hyperedge_index
                add_loop_hyperedge_index = add_self_loop_hyperedges(batch.hyperedge_index, batch.num_nodes)
                batch.hyperedge_index = add_loop_hyperedge_index 
            
            batch = algo_preprocessing(batch,args) 

            if args.method in ['AllSetformer']:
                batch.norm = torch.ones_like(batch.hyperedge_index[0])

            batch = batch.to(device)
            out = model(batch)
            out = F.log_softmax(out, dim=1)  
            pred = out.argmax(dim=1)  
            
            all_preds.append(pred.cpu())  
            all_labels.append(batch.y.cpu())  
    
    y_pred = torch.cat(all_preds).numpy()
    y_true = torch.cat(all_labels).numpy()

    accuracy = accuracy_score(y_true, y_pred)

    macro_f1 = f1_score(y_true, y_pred, average='macro')

    return accuracy, macro_f1

def hg_evaluation_printer(result):
    print(f'Train Acc: {100 * result["acc"][0]:.2f}%, '
        f'Valid Acc: {100 * result["acc"][1]:.2f}%, '
        f'Test  Acc: {100 * result["acc"][2]:.2f}%')
    
    print(f'Train Macro-F1: {100 * result["macro_f1"][0]:.2f}%, '
        f'Valid Macro-F1: {100 * result["macro_f1"][1]:.2f}%, '
        f'Test  Macro-F1: {100 * result["macro_f1"][2]:.2f}%')

def evaluate_edge(model,data,batch_loaders):
    
    train_metrics = eval_edge_train(model,data,batch_loaders)
    
    val_metrics = eval_edge_val_test(model,data,batch_loaders,mode='val')
    
    test_metrics = eval_edge_val_test(model,data,batch_loaders,mode='test')
    
    return train_metrics, val_metrics, test_metrics

def edge_evaluation_printer(train_metrics, val_metrics, test_metrics):

    print('---------Train set-----------')
    print(f"|AUROC| {train_metrics['roc_train']:.4f}")
    print(f"|AP| {train_metrics['ap_train']:.4f}")

    print('---------Valid set-----------')
    print(f"|AUROC| SNS: {val_metrics['roc_sns']:.4f} | MNS:{val_metrics['roc_mns']:.4f} | CNS: {val_metrics['roc_cns']:.4f} | MIX: {val_metrics['roc_mixed']:.4f} | AVG: {val_metrics['roc_average']:.4f}")
    print(f"|AP| SNS: {val_metrics['ap_sns']:.4f} | MNS:{val_metrics['ap_mns']:.4f} | CNS: {val_metrics['ap_cns']:.4f} | MIX: {val_metrics['ap_mixed']:.4f} | AVG: {val_metrics['ap_average']:.4f}")
    
    print('---------Test set-----------')
    print(f"|AUROC| SNS: {test_metrics['roc_sns']:.4f} | MNS:{test_metrics['roc_mns']:.4f} | CNS: {test_metrics['roc_cns']:.4f} | MIX: {test_metrics['roc_mixed']:.4f} | AVG: {test_metrics['roc_average']:.4f}")
    print(f"|AP| SNS: {test_metrics['ap_sns']:.4f} | MNS:{test_metrics['ap_mns']:.4f} | CNS: {test_metrics['ap_cns']:.4f} | MIX: {test_metrics['ap_mixed']:.4f} | AVG: {test_metrics['ap_average']:.4f}")

def aggr_metrics(metrics_dict,result):

    for data_type in result.keys():
        for k,v in result[data_type].items():
            metrics_dict[data_type][k].append(v)
    
    return metrics_dict

def avg_result_printer_edge(metrics_dict):
    
    print('---------Train set-----------')
    train_mean,train_std=[],[]
    for k in metrics_dict['train']:
        train_mean.append(np.mean(metrics_dict['train'][k]))
        train_std.append(np.std(metrics_dict['train'][k]))
    print(f"|AUROC| {train_mean[0]:.4f}+-{train_std[0]:.4f}")
    print(f"|AP| {train_mean[1]:.4f}+-{train_std[1]:.4f}")

    print('---------Valid set-----------')
    val_mean,val_std=[],[]
    for k in metrics_dict['val']:
        val_mean.append(np.mean(metrics_dict['val'][k]))
        val_std.append(np.std(metrics_dict['val'][k]))
    print(f"|AUROC| SNS: {val_mean[0]:.4f}+-{val_std[0]:.4f} | MNS: {val_mean[2]:.4f}+-{val_std[2]:.4f} | CNS: {val_mean[4]:.4f}+-{val_std[4]:.4f} | MIX: {val_mean[6]:.4f}+-{val_std[6]:.4f} | AVG: {val_mean[8]:.4f}+-{val_std[8]:.4f}")
    print(f"|AP| SNS: {val_mean[1]:.4f}+-{val_std[1]:.4f} | MNS: {val_mean[3]:.4f}+-{val_std[3]:.4f} | CNS: {val_mean[5]:.4f}+-{val_std[5]:.4f} | MIX: {val_mean[7]:.4f}+-{val_std[7]:.4f} | AVG: {val_mean[9]:.4f}+-{val_std[9]:.4f}")

    print('---------Test set-----------')
    test_mean,test_std=[],[]
    for k in metrics_dict['test']:
        test_mean.append(np.mean(metrics_dict['test'][k]))
        test_std.append(np.std(metrics_dict['test'][k]))
    print(f"|AUROC| SNS: {test_mean[0]:.4f}+-{test_std[0]:.4f} | MNS: {test_mean[2]:.4f}+-{test_std[2]:.4f} | CNS: {test_mean[4]:.4f}+-{test_std[4]:.4f} | MIX: {test_mean[6]:.4f}+-{test_std[6]:.4f} | AVG: {test_mean[8]:.4f}+-{test_std[8]:.4f}")
    print(f"|AP| SNS: {test_mean[1]:.4f}+-{test_std[1]:.4f} | MNS: {test_mean[3]:.4f}+-{test_std[3]:.4f} | CNS: {test_mean[5]:.4f}+-{test_std[5]:.4f} | MIX: {test_mean[7]:.4f}+-{test_std[7]:.4f} | AVG: {test_mean[9]:.4f}+-{test_std[9]:.4f}")

def evaluate_edge_loader(model, data, dataloader):
    model.eval()
    test_preds, test_labels = [], []

    with torch.no_grad():
        
        while True:
            # 1.message passing
            n,_ = model.encoding(data)

            # 2. candidate scoring for hyperedge in validation/test datasets
            hedges, labels, is_last = dataloader.next() 
            test_preds += model.aggregate(n, hedges, mode='Eval') 
            test_labels.append(labels.detach())
            
            if is_last:
                break

        test_preds = torch.sigmoid(torch.stack(test_preds).squeeze())
        test_labels = torch.cat(test_labels, dim=0)

    return test_preds.tolist(), test_labels.tolist()

def eval_edge_train(model,data,batch_loaders):
    
    train_pred_pos, train_label_pos = evaluate_edge_loader(model, data, batch_loaders['train_pos'])
    train_pred_neg, train_label_neg = evaluate_edge_loader(model, data, batch_loaders['train_neg'])
    
    roc_train = roc_auc_score(np.array(train_label_pos+train_label_neg), np.array(train_pred_pos+train_pred_neg))
    ap_train = average_precision_score(np.array(train_label_pos+train_label_neg), np.array(train_pred_pos+train_pred_neg))
    
    metrics_dict={
        'roc_train':roc_train,
        'ap_train':ap_train
    }
    
    return metrics_dict

def eval_edge_val_test(model,data,batch_loaders,mode='val'):
    
    val_pred_pos, val_label_pos = evaluate_edge_loader(model, data, batch_loaders[f'{mode}_pos'])
    val_pred_sns, val_label_sns = evaluate_edge_loader(model, data, batch_loaders[f'{mode}_neg_sns'])
    val_pred_mns, val_label_mns = evaluate_edge_loader(model, data, batch_loaders[f'{mode}_neg_mns'])
    val_pred_cns, val_label_cns = evaluate_edge_loader(model, data, batch_loaders[f'{mode}_neg_cns'])
    
    # SNS set
    roc_sns = roc_auc_score(np.array(val_label_pos+val_label_sns), np.array(val_pred_pos+val_pred_sns))
    ap_sns = average_precision_score(np.array(val_label_pos+val_label_sns),np.array(val_pred_pos+val_pred_sns))

    # MNS set
    roc_mns = roc_auc_score(np.array(val_label_pos+val_label_mns), np.array(val_pred_pos+val_pred_mns))
    ap_mns = average_precision_score(np.array(val_label_pos+val_label_mns),np.array(val_pred_pos+val_pred_mns))

    # CNS set
    roc_cns = roc_auc_score(np.array(val_label_pos+val_label_cns), np.array(val_pred_pos+val_pred_cns))
    ap_cns = average_precision_score(np.array(val_label_pos+val_label_cns),np.array(val_pred_pos+val_pred_cns))

    # Mixed set
    d = len(val_pred_pos) // 3
    val_label_mixed = val_label_pos + val_label_sns[0:d]+val_label_mns[0:d]+val_label_cns[0:d]
    val_pred_mixed = val_pred_pos + val_pred_sns[0:d]+val_pred_mns[0:d]+val_pred_cns[0:d]
    roc_mixed = roc_auc_score(np.array(val_label_mixed), np.array(val_pred_mixed))
    ap_mixed = average_precision_score(np.array(val_label_mixed),np.array(val_pred_mixed))

    roc_average = (roc_sns+roc_mns+roc_cns+roc_mixed)/4
    ap_average = (ap_sns+ap_mns+ap_cns+ap_mixed)/4

    metrics_dict={
        'roc_sns':roc_sns,
        'ap_sns':ap_sns,
        'roc_mns':roc_mns,
        'ap_mns':ap_mns,
        'roc_cns':roc_cns,
        'ap_cns':ap_cns,
        'roc_mixed':roc_mixed,
        'ap_mixed': ap_mixed,
        'roc_average':roc_average,
        'ap_average':ap_average
    }
    
    return metrics_dict

def fair_metric(pred, labels, sens):

    idx_s0 = sens == 0
    idx_s1 = sens == 1

    idx_s0_y1 = np.bitwise_and(idx_s0, labels == 1)
    idx_s1_y1 = np.bitwise_and(idx_s1, labels == 1)

    parity = abs(sum(pred[idx_s0]) / sum(idx_s0) -
                 sum(pred[idx_s1]) / sum(idx_s1))

    equality = abs(sum(pred[idx_s0_y1]) / sum(idx_s0_y1) -
                   sum(pred[idx_s1_y1]) / sum(idx_s1_y1))
    return parity.item()*100, equality.item()*100

def masked_accuracy(logits: Tensor, labels: Tensor):
    if len(logits) == 0:
        return 0
    pred = torch.argmax(logits, dim=1)
    acc = pred.eq(labels).sum() / len(logits) * 100
    return acc.item()

def accuracy(logits: Tensor, labels: Tensor, masks: dict[Tensor]):
    accs = []
    for mask in masks.values():
        acc = masked_accuracy(logits[mask], labels[mask])
        accs.append(acc)
    return accs

def masked_f1_score(logits: Tensor, labels: Tensor):
    if len(logits) == 0:
        return 0
    pred = torch.argmax(logits, dim=1)
    f1=f1_score(labels.cpu().numpy(), pred.cpu().numpy()) 
    return f1

def f1_scores(logits: Tensor, labels: Tensor, masks: dict[Tensor]):
    f1s = []
    for mask in masks.values():
        f1 = masked_f1_score(logits[mask], labels[mask])
        f1s.append(f1)
    return f1s

def masked_auc_roc(auc_score: Tensor, labels: Tensor):
    if len(auc_score) == 0:
        return 0
    auc_roc=roc_auc_score(labels.cpu().numpy(), auc_score.cpu().numpy())
    return auc_roc

def auc_rocs(logits: Tensor, labels: Tensor, masks: dict[Tensor]):

    probs = F.softmax(logits, dim=1)
    auc_score = probs[:, 1].detach()  
    aucs = []
    for mask in masks.values():
        auc = masked_auc_roc(auc_score[mask], labels[mask])
        aucs.append(auc)
    return aucs

def masked_fairness(logits: Tensor, labels: Tensor, sens):

    if len(logits) == 0:
        return 0
    pred = torch.argmax(logits, dim=1)
    dp,eo = fair_metric(pred.cpu().numpy(), labels.cpu().numpy(), sens.cpu().numpy())
    return dp,eo

def fairness(logits: Tensor, labels: Tensor, masks: dict[Tensor], sens):
    dps,eos = [],[]
    for mask in masks.values():
        dp,eo = masked_fairness(logits[mask], labels[mask], sens[mask])
        dps.append(dp)
        eos.append(eo)
    return dps,eos



import torch
import torch.nn as nn
import torch.nn.functional as F

class PlainUnigencoder(nn.Module):

    def __init__(self, num_features, num_targets, args):
        super(PlainUnigencoder, self).__init__()
        
        self.in_channels=num_features
        self.hidden_channels=args.MLP_hidden
        self.out_channels=num_targets
        self.num_layers=args.All_num_layers
        
        self.lins = nn.ModuleList()
        self.lins.append(nn.Linear(self.in_channels, self.hidden_channels))
        for _ in range(self.num_layers - 2):
            self.lins.append(nn.Linear(self.hidden_channels, self.hidden_channels))
        self.lins.append(nn.Linear(self.hidden_channels, self.out_channels))
        self.cls = nn.Linear(self.out_channels, self.out_channels)
        self.dropout = args.dropout

    def reset_parameters(self):
        for lin in self.lins:
            lin.reset_parameters()

    def forward(self, data):
        x = data.x
        Pv, PvT = data.Pv, data.PvT
        x = torch.spmm(Pv, x)
        for i, lin in enumerate(self.lins[:-1]):
            x = lin(x)
            x = F.relu(x, inplace=True)
            x = F.dropout(x, p=self.dropout, training=self.training)
        x = self.lins[-1](x)
        x = torch.spmm(PvT, x)
        return x,None


import torch
import math
import numpy as np
import scipy.sparse as sp

def symnormalise(M):
    """
    symmetrically normalise sparse matrix

    arguments:
    M: scipy sparse matrix

    returns:
    D^{-1/2} M D^{-1/2} 
    where D is the diagonal node-degree matrix
    """
    
    d = np.array(M.sum(1))
    
    dhi = np.power(d, -1/2).flatten()
    dhi[np.isinf(dhi)] = 0.
    DHI = sp.diags(dhi)    # D half inverse i.e. D^{-1/2}
    
    return (DHI.dot(M)).dot(DHI) 

def ssm2tst(M):
    """
    converts a scipy sparse matrix (ssm) to a torch sparse tensor (tst)

    arguments:
    M: scipy sparse matrix

    returns:
    a torch sparse tensor of M
    """
    
    M = M.tocoo().astype(np.float32)
    
    indices = torch.from_numpy(np.vstack((M.row, M.col))).long()
    values = torch.from_numpy(M.data)
    shape = torch.Size(M.shape)
    
    return torch.sparse.FloatTensor(indices, values, shape)

def row_normalize(mx):
    """Row-normalize sparse matrix"""
    rowsum = np.array(mx.sum(1))
    if np.where(rowsum == 0)[0].shape[0] != 0:
        indices = np.where(rowsum == 0)[0]
        for i in indices:
            rowsum[i] = float('inf')
    r_inv = np.power(rowsum, -1).flatten()
    r_inv[np.isinf(r_inv)] = 0.
    r_mat_inv = sp.diags(r_inv)
    mx = r_mat_inv.dot(mx)
    return mx

def create_coo_from_edge_index(edge_index,is_star=False):

    edge_index_cpu = edge_index.to('cpu')

    row = edge_index_cpu[0].numpy()  
    col = edge_index_cpu[1].numpy()  
    values = np.ones(len(row))  

    num_nodes = row.max() + 1  
    num_hyperedges = col.max() + 1  
    
    if is_star:
        H = sp.coo_matrix((values, (row, col)), shape=(num_hyperedges, num_hyperedges)).astype(np.float32) 
    else:
        H = sp.coo_matrix((values, (row, col)), shape=(num_nodes, num_hyperedges)).astype(np.float32) 

    return H

def glorot(tensor):
    if tensor is not None:
        stdv = math.sqrt(6.0 / (tensor.size(-2) + tensor.size(-1)))
        tensor.data.uniform_(-stdv, stdv)

def zeros(tensor):
    if tensor is not None:
        tensor.data.fill_(0)

def normalize_l2(X):
    """Row-normalize  matrix"""
    rownorm = X.detach().norm(dim=1, keepdim=True)
    scale = rownorm.pow(-1)
    scale[torch.isinf(scale)] = 0.
    X = X * scale
    return X

class SparseMM(torch.autograd.Function):
    """
    Sparse x dense matrix multiplication with autograd support.
    Implementation by Soumith Chintala:
    https://discuss.pytorch.org/t/
    does-pytorch-support-autograd-on-sparse-matrix/6156/7
    """
    @staticmethod
    def forward(ctx, M1, M2):
        ctx.save_for_backward(M1, M2)
        return torch.mm(M1, M2)

    @staticmethod
    def backward(ctx, g):
        M1, M2 = ctx.saved_tensors
        M1 = M1.float()
        M2 = M2.float()
        g1 = g2 = None

        if ctx.needs_input_grad[0]:
            g = g.float()
            g1 = torch.mm(g, M2.t())

        if ctx.needs_input_grad[1]:
            g = g.float()
            g2 = torch.mm(M1.t(), g)

        return g1, g2


import math
import copy
import torch
import torch.nn as nn
import torch.nn.functional as F
import itertools

class TMPHN(nn.Module):
    """
    The mean aggregator for a hypergraph
    """
    def __init__(self, num_features, num_targets, X, neig_dict, args):
        """
        features: a function mapping LongTensor of node ids to FloatTensor of feature values
        structure: a dictionary store the neighbors for all target nodes, 
                i.e., key is target node, value is hyperedges contain the target node
        M: the maximum cardinality of the hypergraph
        """
        super(TMPHN, self).__init__()
  
        self.num_layers = args.All_num_layers
        self.Mlst = args.Mlst
        assert self.num_layers == len(self.Mlst), "The number of layers should be equal to the length of Mlst"
        self.input_dim = num_features
        self.hidden_dim = args.hid_dim
        self.out_dim = num_targets
        self.device = args.device
        
        # initialize TMPHN layers
        features_func = nn.Embedding(X.shape[0], X.shape[1])
        features_func.weight = nn.Parameter(torch.FloatTensor(X.cpu()), requires_grad=False)
        encoders = []
        for l in range(self.num_layers):
            if l == 0: # for the first layer
                agg = TMessagePassing(features_func, neig_dict, self.Mlst[l], args=args)
                enc = Encoder(features_func, X.shape[1], self.hidden_dim, args, aggregator=agg, base_model=None)
            else: # for the subsequent layers
                agg = TMessagePassing(lambda n: encoders[l-1](n), neig_dict, self.Mlst[l], args=args)
                enc = Encoder(lambda n: encoders[l-1](n), encoders[l-1].output_dim, self.hidden_dim, args, aggregator=agg, base_model=encoders[l-1])   
            encoders.append(enc) # add the created encoder to the list
        self.enc = encoders[-1]
      
        # MLP layers for readout
        self.W = nn.Parameter(torch.FloatTensor(self.enc.output_dim, self.out_dim))
        self.b = nn.Parameter(torch.FloatTensor(self.out_dim))
        self.reset_parameters()
        
    def reset_parameters(self):
        std = 1. / math.sqrt(self.W.size(1))
        self.W.data.uniform_(-std, std)
        self.b.data.uniform_(-std, std)

    def forward(self, data):
        X = self.enc(data.target_idx)
        W, b = self.W.to(self.device), self.b.to(self.device)
        Z = torch.matmul(X, W) + b
        return Z,None

class TMessagePassing(nn.Module):
    """
    The mean aggregator for a hypergraph
    """
    def __init__(self, features, structure, M, args):
        """
        features: a function mapping LongTensor of node ids to FloatTensor of feature values
        structure: a dictionary store the neighbors for all target nodes, 
                i.e., key is target node, value is hyperedges contain the target node
        M: the maximum cardinality of the hypergraph
        """
        super(TMessagePassing, self).__init__()

        self.features = features
        self.structure = structure
        self.M = M
        self.device = args.device
    
    def forward(self, target_nodes):
        """
        target_nodes: the node indices we aim to aggregate for
        """
        neigh_feats = torch.stack([self.aggregate_for_one_node(target_node) for target_node in target_nodes], dim=0)
        return neigh_feats
    
    def aggregate_for_one_node(self, target_node):
        
        edges_contain_node = self.structure[int(target_node)]
        
        # load to GPU
        self_feat = self.features(torch.LongTensor([target_node])).squeeze().to(self.device)

        if not edges_contain_node:
            #empty neighbors, no aggregation
            return self_feat
        else:
            edge_embedding = torch.zeros_like(self_feat).to(self.device)
            for edge in edges_contain_node:
                if len(edge) == self.M:
                    edge_embedding += self.aggregate_with_M(edge, target_node)
                elif len(edge) < self.M:
                    edge_embedding += self.aggregate_with_c(edge, target_node)
               
            return edge_embedding

    def aggregate_with_M(self, edge, target_node):
        """
        Same as aggregate_with_c, except this is for edges with cardinality = M
        """
        c = len(edge)
        assert c == self.M, 'the list contain less than M nodes'

        num_perms = math.factorial(len(edge)-1)
        tmp_edge = edge.copy()
        tmp_edge.remove(target_node)

        feat = self.features(torch.LongTensor(tmp_edge)).to(self.device)
            
        to_feats = self.adj_coef(c, target_node) * num_perms * torch.prod((feat), dim=0)
    
        return to_feats

    def aggregate_with_c(self, edge, target_node):
        """
        aggregate high order signals to generate neighboring embedding for the target node
        edge: an edge list that the target_node lies in
        target_node: target node index
        features: feature matrix in N x d
        M: maximum cardinality of edges.
        output: edge embeddings for the target node in 1 x d.
        """
        c = len(edge)
        assert c < self.M, 'the list contain exactly or more than M nodes'
        all_comb = [list(t) for t in itertools.combinations_with_replacement(edge, self.M)] #all possible combs to fill in length-M list
        val_comb = list(filter(lambda comb: set(comb) == set(edge), all_comb)) #each node must appear at least once
        
        tmp_comb = copy.deepcopy(val_comb)
        for comb in tmp_comb:
            comb.remove(target_node)
        
        num_perms = torch.Tensor([len(list(set(itertools.permutations(comb)))) for comb in tmp_comb])

        #cross multiply features

        high_order_signal = torch.stack([torch.prod(self.features(torch.LongTensor(comb)), dim=0) for comb in tmp_comb])
        num_perms = num_perms.to(self.device)
        high_order_signal = high_order_signal.to(self.device)
        agg = torch.matmul(num_perms, high_order_signal)

        agg_with_adj = self.adj_coef(c, target_node) * agg

        return agg_with_adj
    
    def adj_coef(self, c, node):
        """
        compute the adjacency coefficient for hyperedges.
        c: cardinality of hyperedge.
        M: maximum cardinality of hyperedge.
        alpha: the sum of multinomial coefficients over positive integer.
        """
        alpha = 0
        for i in range(c):
            alpha += ((-1) ** i) * math.comb(c, i) * ((c - i) ** self.M)
        a = c / alpha
        degree = len(self.structure[int(node)])
        return a/degree
    
class Encoder(nn.Module):
    """
    Encodes nodes features (mapping features to different dimension)
    """
    def __init__(self, features, input_dim, output_dim, args, aggregator, base_model=None):
        """
        features: a function maps node indices to corresponding features
        feature_dim: input feature dimension
        embed_dim: the output feature dimension
        """
        super(Encoder, self).__init__()
        self.features = features
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.combine = args.combine
        self.aggregator = aggregator
        self.device = args.device

        if base_model != None:
            self.base_model = base_model

        self.W = nn.Parameter(
                torch.FloatTensor(self.input_dim if self.combine == 'sum' else 2 * self.input_dim, self.output_dim))
        self.b = nn.Parameter(torch.FloatTensor(self.output_dim))
        self.reset_parameters()

    def reset_parameters(self):
        std = 1. / math.sqrt(self.W.size(1))
        self.W.data.uniform_(-std, std)
        self.b.data.uniform_(-std, std)
    
    def forward(self, nodes):

        neigh_feats = self.aggregator.forward(nodes) # A*X
        self_feats = self.features(torch.LongTensor(nodes)).to(self.device) # X
        if self.combine == 'concat':
            combined = torch.cat([self_feats, neigh_feats], dim=1) #concatenate self features
        else:
            combined = self_feats + neigh_feats # sum self features
        
        W, b = self.W.to(self.device), self.b.to(self.device)
        
        AXW = torch.mm(combined, W)
        
        y = AXW + b

        output = F.relu(y)
        return output #output is in dimension (num_of_nodes, embed_dim)

    def __repr__(self):
        return self.__class__.__name__ + ' (' \
               + str(self.input_dim) + ' -> ' \
               + str(self.output_dim) + ')'




import copy
import torch, math, numpy as np, scipy.sparse as sp
import torch.nn as nn, torch.nn.functional as F
from torch.autograd import Variable
from dhgbench.lib_models.HNN.utils import SparseMM
from dhgbench.lib_models.HNN.preprocessing import get_HyperGCN_He_dict,update
from dhgbench.lib_models.HNN.utils import symnormalise,ssm2tst
from collections import defaultdict
from tqdm import tqdm

class HyperGCNConv(nn.Module):
    """
    Simple GCN layer, similar to https://arxiv.org/abs/1609.02907
    """
    def __init__(self, a, b, reapproximate=True):
        super(HyperGCNConv, self).__init__()
        self.a, self.b = a, b
        self.reapproximate = reapproximate

        self.W = nn.Parameter(torch.FloatTensor(a, b))
        self.bias = nn.Parameter(torch.FloatTensor(b))
        self.reset_parameters()
        
    def reset_parameters(self):
        std = 1. / math.sqrt(self.W.size(1))
        self.W.data.uniform_(-std, std)
        self.bias.data.uniform_(-std, std)

    def forward(self, structure, H, m=True):
#         ipdb.set_trace()
        W, b = self.W, self.bias
        HW = torch.mm(H, W)

        if self.reapproximate:
            n, X = H.shape[0], HW.cpu().detach().numpy()
            A = Laplacian(n, structure, X, m)
        else: A = structure

        A = A.to(H.device)
        A = Variable(A)

        AHW = SparseMM.apply(A, HW)     
        return AHW + b

    def __repr__(self):
        return self.__class__.__name__ + ' (' \
               + str(self.a) + ' -> ' \
               + str(self.b) + ')'

class HyperGCN(nn.Module):
    def __init__(self, num_features, num_targets, args):
        """
        d: initial node-feature dimension
        h: number of hidden units
        c: number of classes
        """
        super(HyperGCN, self).__init__()

        self.fast = args.HyperGCN_fast
        self.structure, self.mediator = None, args.HyperGCN_mediators
        self.dropout, self.num_layers = args.dropout, args.All_num_layers
        self.hidden_dim = args.MLP_hidden

        self.convs = nn.ModuleList()
        if self.num_layers == 1:
            self.convs.append(HyperGCNConv(
                num_features, num_targets, (not self.fast)))
        else:
            self.convs.append(HyperGCNConv(
                num_features, self.hidden_dim, (not self.fast)))
            for _ in range(self.num_layers-2):
                self.convs.append(HyperGCNConv(
                self.hidden_dim, self.hidden_dim, (not self.fast)))
            self.convs.append(HyperGCNConv(
                self.hidden_dim, num_targets, (not self.fast)))

    def reset_parameters(self):
        for layer in self.convs:
            layer.reset_parameters()

    def forward(self, data):
        """
        an l-layer GCN
        """
        if self.structure is None:
            print('Precomputing ...')
            data_copy = copy.deepcopy(data)
            data_copy.to('cpu')
            He_dict = get_HyperGCN_He_dict(data_copy) 
            
            if self.fast:
                self.structure = Laplacian(V=data_copy.x.shape[0], E=He_dict, X=data_copy.x, m=self.mediator)
            else:
                self.structure = He_dict

        H = data.x
        
        for i, conv in enumerate(self.convs):
            H = F.relu(conv(self.structure, H, self.mediator))
            if i < self.num_layers - 1:
                H = F.dropout(H, self.dropout, training=self.training)

        return H,None 

def adjacency(edges, weights, n):
    """
    computes an sparse adjacency matrix

    arguments:
    edges: list of pairs
    weights: dictionary of edge weights (key: tuple representing edge, value: weight on the edge)
    n: number of nodes

    returns: a scipy.sparse adjacency matrix with unit weight self loops for edges with the given weights
    """
    
    dictionary = {tuple(item): index for index, item in enumerate(edges)}
    edges = [list(itm) for itm in dictionary.keys()]   
    organised = []

    for e in edges:
        i,j = e[0],e[1]
        w = weights[(i,j)]
        organised.append(w)

    edges, weights = np.array(edges), np.array(organised)
    adj = sp.coo_matrix((weights, (edges[:, 0], edges[:, 1])), shape=(n, n), dtype=np.float32)
    adj = adj + sp.eye(n)

    A = symnormalise(sp.csr_matrix(adj, dtype=np.float32))
    A = ssm2tst(A)
    return A

def Laplacian(V, E, X, m):
    """
    Approximates hypergraph Laplacian using clique expansion or mediator method.

    Args:
        V (int): number of vertices
        E (dict): hyperedge dictionary {hid: list of nodes}
        X (ndarray): node features
        m (bool): use mediator mode or not

    Returns:
        A (torch.sparse.Tensor): normalized adjacency matrix
    """
    edges = []
    weights = defaultdict(float)
    rv = np.random.rand(X.shape[1])
    X_proj = X @ rv  # shape: (n_nodes,)

    for hyperedge in tqdm(E.values()):

        hyperedge = list(hyperedge)
        proj = X_proj[hyperedge]
        s, i = np.argmax(proj), np.argmin(proj)
        Se, Ie = hyperedge[s], hyperedge[i]

        if m:
            c = max(2 * len(hyperedge) - 3, 1)

            edges.extend([[Se, Ie], [Ie, Se]])
            weights[(Se, Ie)] += 1.0 / c
            weights[(Ie, Se)] += 1.0 / c

            for mediator in hyperedge:
                if mediator != Se and mediator != Ie:
                    for u, v in [(Se, mediator), (Ie, mediator), (mediator, Se), (mediator, Ie)]:
                        edges.append([u, v])
                        weights[(u, v)] += 1.0 / c
        else:
            e = len(hyperedge)
            edges.extend([[Se, Ie], [Ie, Se]])
            weights[(Se, Ie)] += 1.0 / e
            weights[(Ie, Se)] += 1.0 / e

    return adjacency(edges, weights, V)

def Laplacian_old(V, E, X, m):
    """
    approximates the E defined by the E Laplacian with/without mediators

    arguments:
    V: number of vertices
    E: dictionary of hyperedges (key: hyperedge, value: list/set of hypernodes)
    X: features on the vertices
    m: True gives Laplacian with mediators, while False gives without

    A: adjacency matrix of the graph approximation
    returns: 
    updated data with 'graph' as a key and its value the approximated hypergraph 
    """
    
    edges, weights = [], {}
    rv = np.random.rand(X.shape[1])

    for k in E.keys():
        hyperedge = list(E[k])
        
        p = np.dot(X[hyperedge], rv)   #projection onto a random vector rv
        s, i = np.argmax(p), np.argmin(p)
        Se, Ie = hyperedge[s], hyperedge[i]

        # two stars with mediators
        # c = 2*len(hyperedge) - 3   
        c = max(2*len(hyperedge) - 3,1)   # normalisation constant
        if m:
            
            # connect the supremum (Se) with the infimum (Ie)
            edges.extend([[Se, Ie], [Ie, Se]])
            
            if (Se,Ie) not in weights:
                weights[(Se,Ie)] = 0
            weights[(Se,Ie)] += float(1/c)

            if (Ie,Se) not in weights:
                weights[(Ie,Se)] = 0
            weights[(Ie,Se)] += float(1/c)
            
            # connect the supremum (Se) and the infimum (Ie) with each mediator
            for mediator in hyperedge:
                if mediator != Se and mediator != Ie:
                    edges.extend([[Se,mediator], [Ie,mediator], [mediator,Se], [mediator,Ie]])
                    weights = update(Se, Ie, mediator, weights, c)
        else:
            edges.extend([[Se,Ie], [Ie,Se]])
            e = len(hyperedge)
            
            if (Se,Ie) not in weights:
                weights[(Se,Ie)] = 0
            weights[(Se,Ie)] += float(1/e)

            if (Ie,Se) not in weights:
                weights[(Ie,Se)] = 0
            weights[(Ie,Se)] += float(1/e)    
    
    return adjacency(edges, weights, V)



import time
import copy
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from tqdm import tqdm,trange
from dhgbench.lib_utils.utils import mask_to_index,relabel_hyperedge_index,add_self_loop_hyperedges
from dhgbench.lib_utils.metrics import evaluate,evaluate_edge,edge_evaluation_printer,hg_evaluation_printer,evaluate_hypegraph
from dhgbench.lib_models import _semi_methods_
from sklearn.utils.class_weight import compute_class_weight
from dhgbench.lib_models.HNN.preprocessing import algo_preprocessing

class Trainer:
    
    def __init__(self, args, **kwargs):
        """
        Training pipline for different kinds of models
        """
        self.args = args
        self.device=args.device
    
    def semi_hg_cls_training(self,model,data,batch_loaders,args):

        device = args.device

        if args.use_weighted_loss:
            class_weights = compute_class_weight(class_weight="balanced", classes=np.unique(data.y.numpy()), y=data.y.numpy())
            class_weights = torch.tensor(class_weights, dtype=torch.float)
        else:
            class_weights = None  
        
        criterion = nn.NLLLoss(weight=class_weights.to(device) if args.use_weighted_loss else None)
        
        start_time = time.time()
        model.reset_parameters()
        optimizer = torch.optim.Adam(model.parameters(), lr=args.lr, weight_decay=args.wd)
        
        train_loader, val_loader, test_loader = batch_loaders['train'], batch_loaders['val'], batch_loaders['test']
        
        best_val_acc = -1 
        best_model = None

        for epoch in tqdm(range(args.epochs)):
            
            # Training part
            model.train()
            optimizer.zero_grad()
            
            total_loss = 0
            
            for batch in train_loader:
                
                if args.method in ['HyperND','TFHNN']:
                    model.encoder.cache=None 
                elif args.method in ['HyperGCN']:
                    model.encoder.structure=None
                elif args.method in ['SheafHyperGNN']:
                    model.encoder.hyperedge_attr=None 
                    
                if args.method in ['HyperND','TFHNN','HyperGCN','SheafHyperGNN']:
                    reindex_hyperedge_index,_ = relabel_hyperedge_index(batch.hyperedge_index)
                    batch.hyperedge_index = reindex_hyperedge_index
                    add_loop_hyperedge_index = add_self_loop_hyperedges(batch.hyperedge_index, batch.num_nodes)
                    batch.hyperedge_index = add_loop_hyperedge_index 
                
                batch = algo_preprocessing(batch,args) 
                
                if args.method in ['AllSetformer']:
                    batch.norm = torch.ones_like(batch.hyperedge_index[0])

                batch = batch.to(device)
                optimizer.zero_grad()
                out = model(batch)
                out = F.log_softmax(out, dim=1)
                loss = criterion(out, batch.y)
                loss.backward()
                optimizer.step()
                total_loss += loss.item()
            
            avg_loss = total_loss / len(train_loader) 
            
            if (epoch+1) % args.display_step == 0:
                
                result = evaluate_hypegraph(model,batch_loaders,args) 
                
                if result['acc'][1]>=best_val_acc:
                    best_val_acc = result['acc'][1]
                    best_model = copy.deepcopy(model)

                print(f'Epoch: {epoch+1:02d}, Training loss: {avg_loss:.4f}')
                hg_evaluation_printer(result)
                
        end_time = time.time()
        print(f'Training Time: {end_time-start_time:.2f}')
        
        if args.early_stop:
            
            return best_model
        else:
            return model

    def semi_edge_pred_training(self,model,data,batch_loaders,args):

        start_time = time.time()
        model.reset_parameters()
        optimizer = torch.optim.Adam(model.parameters(), lr=args.lr, weight_decay=args.wd)

        train_pos_loader = batch_loaders['train_pos']
        train_neg_loader = batch_loaders['train_neg']

        best_val_auc = -1 
        best_model = None

        for epoch in tqdm(range(args.epochs)):
            
            # Training part
            model.train()
            total_loss = 0.0
            
            while True :
                
                optimizer.zero_grad()
                
                n,_ = model.encoding(data) 
                
                pos_hedges, pos_labels, is_last = train_pos_loader.next()
                neg_hedges, neg_labels, is_last = train_neg_loader.next()

                pos_preds = model.aggregate(n, pos_hedges, mode='Train') 
                neg_preds = model.aggregate(n, neg_hedges, mode='Train') 
                pos_preds,neg_preds = torch.sigmoid(pos_preds),torch.sigmoid(neg_preds)
                
                d_real_loss = F.binary_cross_entropy_with_logits(pos_preds, pos_labels) 
                d_fake_loss = F.binary_cross_entropy_with_logits(neg_preds, neg_labels) 
                train_loss = d_real_loss + d_fake_loss
                
                train_loss.backward()
                optimizer.step()

                total_loss += train_loss.item()
                
                if is_last :
                    break 
            
            if (epoch+1) % args.display_step == 0:
                
                print(f'Epoch: {epoch+1:02d}, Training loss: {total_loss:.4f}')
                
                train_metrics, val_metrics, test_metrics = evaluate_edge(model,data,batch_loaders)

                if val_metrics['roc_average']>=best_val_auc:

                    best_val_auc = val_metrics['roc_average']
                    best_model = copy.deepcopy(model)

                edge_evaluation_printer(train_metrics, val_metrics, test_metrics)
        
        end_time = time.time()
        print(f'Training Time: {end_time-start_time:.2f}')

        if args.early_stop:
            return best_model
        else:
            return model
    
    def semi_node_cls_training(self,model,data,masks,args):
        
        criterion = nn.NLLLoss()
        
        model.train()
        
        ### Training loop ###
        start_time = time.time()
        model.reset_parameters()
        
        if args.method == 'UniGCNII':
            optimizer = torch.optim.Adam([
                dict(params=model.reg_params, weight_decay=0.01),
                dict(params=model.non_reg_params, weight_decay=5e-4)
            ], lr=0.01)
        else:
            optimizer = torch.optim.Adam(model.parameters(), lr=args.lr, weight_decay=args.wd)
        
        if args.method == 'HJRL':
            pos_weight_H = float(data.H_ini.shape[0] * data.H_ini.shape[0] - data.H_ini.sum()) / data.H_ini.sum()
            if torch.isinf(pos_weight_H):
                pos_weight_H = torch.FloatTensor(args.pos_weight_thresh)
        elif args.method == 'TMPHN':
            train_idx = mask_to_index(masks['train'].to('cpu'),data.x.shape[0])
            data.target_idx = train_idx

        #for epoch in tqdm(range(args.epochs)):
        for epoch in trange(args.epochs):
        # for epoch in range(args.epochs):
            # Training part
            model.train()
            optimizer.zero_grad()
            if args.method == 'HJRL':
                node_embed,edge_embed = model(data)
                # 1. node classification loss
                node_embed = F.log_softmax(node_embed, dim=1)
                loss_1 = criterion(node_embed[masks['train']],data.y[masks['train']])
                if args.gamma == 0:
                    loss_2 = 0
                else:
                    # 2. reconstruction loss
                    recovered_H = torch.mm(node_embed, edge_embed.t())
                    recovered_H = torch.sigmoid(recovered_H)
                    if args.sample_ratio:
                        sampled_row = np.random.choice(recovered_H.shape[0], int(recovered_H.shape[0] * args.sample_ratio), replace=False)
                        loss_2 = F.binary_cross_entropy_with_logits(recovered_H[sampled_row], data.H_ini[sampled_row], pos_weight=pos_weight_H)
                    else:
                        loss_2 = F.binary_cross_entropy_with_logits(recovered_H, data.H_ini, pos_weight=pos_weight_H)
                loss = loss_1 + args.gamma * loss_2
                loss.backward()
                optimizer.step()
            else:
                out,_ = model(data) 
                out = F.log_softmax(out, dim=1)
                if args.method == 'TMPHN':
                    loss = criterion(out,data.y[masks['train']])
                else:
                    loss = criterion(out[masks['train']],data.y[masks['train']])
                loss.backward()
                if args.clip_grad:
                    torch.nn.utils.clip_grad_norm_(model.parameters(), args.clip_thresh)
                optimizer.step()

            if (epoch+1) % args.display_step == 0:
                
                if args.method == 'TMPHN':
                    data.target_idx=data.all_idx
                    
                result = evaluate(model, data, masks) 
                
                print(f'Epoch: {epoch+1:02d}, '
                    f'Train Acc: {100 * result[0]:.2f}%, '
                    f'Valid Acc: {100 * result[1]:.2f}%, '
                    f'Test  Acc: {100 * result[2]:.2f}%')

                if args.method == 'TMPHN':
                    data.target_idx=train_idx
                
        end_time = time.time()
        self.train_time = end_time-start_time
        print(f'Training Time: {end_time-start_time:.2f}')
        
        if args.method == 'TMPHN':
            data.target_idx=data.all_idx 
   
    def training(self,model,data,args,seed_split=None,task_type='node_cls'):
        
        if self.args.method in _semi_methods_:
            if task_type == 'node_cls':
                self.semi_node_cls_training(model,data,seed_split,args)
            elif task_type == 'edge_pred':
                model = self.semi_edge_pred_training(model,data,seed_split,args)
                return model
            elif task_type == 'hg_cls':
                model = self.semi_hg_cls_training(model,data,seed_split,args)
                return model
            else:
                raise NotImplementedError


import time
import os
import os.path as osp
import random
import copy
import torch
import torch.nn.functional as F
import torch_sparse
import torch_scatter
import numpy as np
from collections import defaultdict
import dask
from dask.diagnostics import ProgressBar
from dask import delayed
import scipy.sparse as sp 
from dhgbench.lib_models.HNN.utils import symnormalise,ssm2tst,row_normalize,create_coo_from_edge_index
from typing import Optional
from tqdm import tqdm
import os 
from torch_geometric.utils import remove_self_loops

def algo_preprocessing(data,args):

    if args.method =='LEGCN':
        data = legcn_preprocessing(data,args)  # line_expansion
    elif args.method in ['HNHN']:
        data = generate_HNHN_norm(data,args)
    elif args.method in ['PhenomNNS','PhenomNN']:
        data = phenomNN_preprocessing(data,args)
    elif args.method == 'HJRL':
        data = hjrl_preprocessing(data,args)
    elif args.method == 'TMPHN':
        data = tmphn_preprocessing(data,args) 
    elif args.method == 'DPHGNN':
        data = dphgnn_preprocessing(data,args) 
    elif args.method == 'EHNN':
        data = ehnn_preprocessing(data,args)
    elif args.method in ['PlainUnigencoder']:
        data =uni_expansion(data,args)
    elif args.method == 'HyperGT':
        data = hypergt_preprocessing(data,args)
    elif args.method in ['CEGCN','CEGAT']:
        data = cegnn_preprocessing(data,args)
    else:
        pass
    
    return data

# ------------ Hypergraph Laplacian ----------------------

def hgnn_laplacian(H,type='full'):

    if type=="rw":
        # 1. H x B x H^{T}
        colsum = np.array(H.sum(0),dtype='float')
        r_inv_sqrt = np.power(colsum, -1).flatten()
        r_inv_sqrt[np.isinf(r_inv_sqrt)] = 0.
        B = sp.diags(r_inv_sqrt)

        A_beta = H.dot(B).dot(H.transpose())
        I = sp.eye(A_beta.shape[0])
        A_beta+=I
        
        rowsum = np.array(A_beta.sum(1),dtype='float')
        r_inv_sqrt = np.power(rowsum, -1.0).flatten()
        r_inv_sqrt[np.isinf(r_inv_sqrt)] = 0.
        D = sp.diags(r_inv_sqrt)
        
        A_beta = D.dot(A_beta)
        
        L = I-A_beta

    else:
        # 1. H x B x H^{T}
        colsum = np.array(H.sum(0),dtype='float')
        r_inv_sqrt = np.power(colsum, -1).flatten()
        r_inv_sqrt[np.isinf(r_inv_sqrt)] = 0.
        B = sp.diags(r_inv_sqrt)

        A_beta = H.dot(B).dot(H.transpose())
        I = sp.eye(A_beta.shape[0])
        A_beta+=I

        # 2. D^{-1/2} x H x B x H^{T} x D^{1/2}
        rowsum = np.array(A_beta.sum(1),dtype='float')
        r_inv_sqrt = np.power(rowsum, -0.5).flatten()
        r_inv_sqrt[np.isinf(r_inv_sqrt)] = 0.
        D = sp.diags(r_inv_sqrt)

        A_beta = D.dot(A_beta).dot(D)
        
        if type=='sym':
            I = sp.eye(A_beta.shape[0])
            L=I-A_beta
        elif type=='full':
            L=A_beta
    
    return L

def Lapalacian_generator(data,args):
    H = create_coo_from_edge_index(data.hyperedge_index) 
    L_hgnn,L_sym,L_rw = hgnn_laplacian(H,type='full'),hgnn_laplacian(H,type='sym'),hgnn_laplacian(H,type='rw')
    return ssm2tst(L_hgnn).to(args.device),ssm2tst(L_sym).to(args.device),ssm2tst(L_rw).to(args.device)

# ------------ clique expansion methods ----------------------

def clique_expansion(data):
    H = create_coo_from_edge_index(data.hyperedge_index)
    clique_unorm=H.dot(H.transpose())
    clique_unorm.sum_duplicates() 
    src_idx,dst_idx=clique_unorm.nonzero()
    clique_edge_index = torch.tensor(np.vstack((src_idx, dst_idx)), dtype=torch.long)
    return clique_edge_index

# ------------ star expansion methods ----------------------

def star_expansion(data, num_nodes=None, num_hyperedges=None):

    nodes, hyperedges = data.hyperedge_index  

    if num_nodes is None:
        num_nodes = nodes.max().item() + 1  
    
    if num_hyperedges is None:
        num_hyperedges = hyperedges.max().item() + 1  

    hyperedge_offset = num_nodes  
    hyperedge_nodes = hyperedges + hyperedge_offset

    edge_index = torch.stack([nodes, hyperedge_nodes], dim=0)

    v_mask = torch.cat([torch.ones(num_nodes, dtype=torch.bool), torch.zeros(num_hyperedges, dtype=torch.bool)])

    return edge_index, v_mask

# ------------- CEGNN preprocessing utils ----------------------

def cegnn_preprocessing(data,args):
    data.clique_edge_index = clique_expansion(data).to(args.device)
    return data

# ------------- DPHGNN preprocessing utils ----------------------

def dphgnn_preprocessing(data,args):
    # 1. clique expansion
    A_cg,L_cg = clique_preprocessing(data,args)
    
    # 2. hypergcn expansion
    A_hyp,L_hyp = hypergcn_preprocessing(data,args)
    
    # 3. star expansion
    A_sg,L_sg,v_mask = star_preprocessing(data,args)
    
    data.A_cg,data.A_hyp,data.A_sg,data.L_cg,data.L_hyp,data.L_sg,data.v_mask = A_cg,A_hyp,A_sg,L_cg,L_hyp,L_sg,v_mask
    
    data.L_hgnn,data.L_sym,data.L_rw = Lapalacian_generator(data,args)
    
    return data 

def gen_laplacian(A):
    rowsum = np.array(A.sum(1),dtype='float').flatten()
    D = sp.diags(rowsum)
    L=D-A
    return L

def star_preprocessing(data,args):
    star_graph,v_mask = star_expansion(data)
    star_graph_coo = create_coo_from_edge_index(star_graph,is_star=True) 
    
    A_sg=row_normalize(star_graph_coo)
    I = sp.eye(A_sg.shape[0])
    A_sg += I 
    
    L_sg = gen_laplacian(star_graph_coo)
    
    return ssm2tst(A_sg).to(args.device), ssm2tst(L_sg).to(args.device),v_mask

def clique_preprocessing(data,args):
    
    clique_graph = clique_expansion(data) 
    clique_graph_coo = create_coo_from_edge_index(clique_graph) 
    
    A_cg=row_normalize(clique_graph_coo)
    I = sp.eye(A_cg.shape[0])
    A_cg += I # (I+D_{v}^{-1}A_{c})
    
    L_cg = gen_laplacian(clique_graph_coo)
    return ssm2tst(A_cg).to(args.device),ssm2tst(L_cg).to(args.device)

def hypergcn_preprocessing(data,args):

    data_copy = copy.deepcopy(data)
    data_copy.to('cpu')
    He_dict = get_HyperGCN_He_dict(data_copy)
    A = hypergcn_expansion(V=data_copy.x.shape[0], E=He_dict, X=data_copy.x, m=args.mediator) 
    A_hyp = symnormalise(A) 
    L_hyp = gen_laplacian(A)
    
    return ssm2tst(A_hyp).to(args.device),ssm2tst(L_hyp).to(args.device)

def hypergcn_expansion(V, E, X, m):
    """
    Approximates hypergraph Laplacian using clique expansion or mediator method.

    Args:
        V (int): number of vertices
        E (dict): hyperedge dictionary {hid: list of nodes}
        X (ndarray): node features
        m (bool): use mediator mode or not

    Returns:
        A (torch.sparse.Tensor): normalized adjacency matrix
    """
    edges = []
    weights = defaultdict(float)
    rv = np.random.rand(X.shape[1])
    X_proj = X @ rv  # shape: (n_nodes,)

    for hyperedge in E.values():

        hyperedge = list(hyperedge)
        proj = X_proj[hyperedge]
        s, i = np.argmax(proj), np.argmin(proj)
        Se, Ie = hyperedge[s], hyperedge[i]

        if m:
            c = max(2 * len(hyperedge) - 3, 1)

            edges.extend([[Se, Ie], [Ie, Se]])
            weights[(Se, Ie)] += 1.0 / c
            weights[(Ie, Se)] += 1.0 / c

            for mediator in hyperedge:
                if mediator != Se and mediator != Ie:
                    for u, v in [(Se, mediator), (Ie, mediator), (mediator, Se), (mediator, Ie)]:
                        edges.append([u, v])
                        weights[(u, v)] += 1.0 / c
        else:
            e = len(hyperedge)
            edges.extend([[Se, Ie], [Ie, Se]])
            weights[(Se, Ie)] += 1.0 / e
            weights[(Ie, Se)] += 1.0 / e

    return adjacency(edges, weights, V)

def adjacency(edges, weights, n):
    """
    computes an sparse adjacency matrix

    arguments:
    edges: list of pairs
    weights: dictionary of edge weights (key: tuple representing edge, value: weight on the edge)
    n: number of nodes

    returns: a scipy.sparse adjacency matrix with unit weight self loops for edges with the given weights
    """
    
    dictionary = {tuple(item): index for index, item in enumerate(edges)}
    edges = [list(itm) for itm in dictionary.keys()]   
    organised = []

    for e in edges:
        i,j = e[0],e[1]
        w = weights[(i,j)]
        organised.append(w)

    edges, weights = np.array(edges), np.array(organised)
    adj = sp.coo_matrix((weights, (edges[:, 0], edges[:, 1])), shape=(n, n), dtype=np.float32)
    adj = adj + sp.eye(n)

    A = sp.csr_matrix(adj, dtype=np.float32)
    return A

# ------------- PhenomNN preprocessing utils ----------------------

def phenomNN_expansion(H,args,norm_type='full'):

    if norm_type=="full":

        # 1. H x B x H^{T}
        colsum = np.array(H.sum(0),dtype='float')
        r_inv_sqrt = np.power(colsum, -1).flatten()
        r_inv_sqrt[np.isinf(r_inv_sqrt)] = 0.
        B = sp.diags(r_inv_sqrt)

        A_beta = H.dot(B).dot(H.transpose())
        I = sp.eye(A_beta.shape[0])
        A_beta += I

        # 2. D^{-1/2} x H x B x H^{T} x D^{1/2}
        rowsum = np.array(A_beta.sum(1),dtype='float')
        r_inv_sqrt = np.power(rowsum, -0.5).flatten()
        r_inv_sqrt[np.isinf(r_inv_sqrt)] = 0.
        D = sp.diags(r_inv_sqrt)

        A_beta = D.dot(A_beta).dot(D)

        rowsum_norm = np.array(A_beta.sum(1))
        rowsum_norm_flat = rowsum_norm.flatten()
        D_beta = sp.diags(rowsum_norm_flat)

    elif norm_type=="node":

        A_beta=H.dot(H.transpose())
        I = sp.eye(A_beta.shape[0])   
        A_beta+=I
        
        rowsum = np.array(A_beta.sum(1),dtype='float')
        r_inv_sqrt = np.power(rowsum, -0.5).flatten()
        r_inv_sqrt[np.isinf(r_inv_sqrt)] = 0.
        D = sp.diags(r_inv_sqrt)

        A_beta = D.dot(A_beta).dot(D)

        rowsum_norm = np.array(A_beta.sum(1))
        rowsum_norm_flat = rowsum_norm.flatten()
        D_beta = sp.diags(rowsum_norm_flat)

    return ssm2tst(A_beta).to(args.device),ssm2tst(D_beta).to(args.device),ssm2tst(I).to(args.device) 

def phenomNN_preprocessing(data,args):

    H = create_coo_from_edge_index(data.hyperedge_index)

    if args.method == 'PhenomNN':
        A_beta,D_beta,I=phenomNN_expansion(H,args,norm_type="node")
        A_gamma,D_gamma,_=phenomNN_expansion(H,args,norm_type="full")
        H_expansion=[A_beta,A_gamma]
        G=[D_beta,D_gamma,I]
    else:
        A_beta,D_beta,I=phenomNN_expansion(H,args,norm_type="node")
        A_gamma,D_gamma,_=phenomNN_expansion(H,args,norm_type="full")
        A= args.lam0 * A_beta +args.lam1 *  A_gamma
        D=args.lam0 * D_beta +args.lam1 *  D_gamma
        H_expansion=A 
        G=[D,I]

    data.adj, data.G = H_expansion, G
    
    return data

# ------------- HNHN preprocessing utils ----------------------

def generate_HNHN_norm(data, args):
    """
    :param H: hypergraph incidence matrix H
    :param variable_weight: whether the weight of hyperedge is variable
    :return: G
    """

    edge_index=copy.deepcopy(data.hyperedge_index)
    edge_index = edge_index.to('cpu')

    # Construct incidence matrix H of size (num_nodes, num_hyperedges) from edge_index = [V;E]
    ones = torch.ones(edge_index.shape[1], device=edge_index.device)

    alpha = args.HNHN_alpha
    beta = args.HNHN_beta

    # the degree of the node
    DV = torch_scatter.scatter_add(ones, edge_index[0], dim=0)
    # the degree of the hyperedge
    DE = torch_scatter.scatter_add(ones, edge_index[1], dim=0)

    # alpha part
    D_e_alpha = DE ** alpha
    D_e_alpha[D_e_alpha == float("inf")] = 0 
    D_v_alpha = torch_scatter.scatter_add(DE[edge_index[1]], edge_index[0], dim=0)

    # beta part
    D_v_beta = DV ** beta
    D_v_beta[D_v_beta == float("inf")] = 0 
    D_e_beta = torch_scatter.scatter_add(DV[edge_index[0]], edge_index[1], dim=0)

    D_v_alpha_inv = 1.0 / D_v_alpha
    D_v_alpha_inv[D_v_alpha_inv == float("inf")] = 0

    D_e_beta_inv = 1.0 / D_e_beta
    D_e_beta_inv[D_e_beta_inv == float("inf")] = 0

    data.D_e_alpha,data.D_v_alpha_inv,data.D_v_beta,data.D_e_beta_inv = D_e_alpha.float().to(args.device),D_v_alpha_inv.float().to(args.device),D_v_beta.float().to(args.device),D_e_beta_inv.float().to(args.device)
    
    return data

# ------------- Unigencoder preprocessing utils ----------------------------

def uni_expansion(data, args):
    
    threshold, norm_type, init_val, init_type = args.threshold, args.norm_type, args.init_val, args.init_type
    
    V, E = data.hyperedge_index
    # print(data.edge_index.shape)
    edge_dict = {}
    for i in range(data.hyperedge_index.shape[1]):
        if E[i].item() not in edge_dict:
            edge_dict[E[i].item()] = []
        edge_dict[E[i].item()].append(V[i].item())

    N_vertex = V.max() + 1
    N_hyperedge = data.hyperedge_index.shape[1]
    N_hy = len(edge_dict)
    self_set = set()
    # print(len(edge_dict), N_hyperedge)
    for key, val in edge_dict.items():
        if len(val) == 1:
            self_set.add(val[0])
    # print(len(self_set))
    if len(self_set) < N_vertex:
        # print(len(self_set))
        count = 0
        for i in range(N_vertex):
            if i not in self_set:
                edge_dict[N_hy + count] = []
                edge_dict[N_hy+count].append(i)
        count+=1

    node_num_egde = {}
    for key, val in edge_dict.items():
        for v in val:
            if v not in node_num_egde:
                node_num_egde[v] = 0
            else:
                node_num_egde[v] += 1

    pv_rows = []
    pv_cols = []

    n_count = 0
    if threshold >= -1.0:
        def cosine_similarity_dense_small(x):
            norm = F.normalize(x, p=2., dim=-1)  # [N, out_channels]
            sim = norm.mm(norm.t())
            return sim
        sim = cosine_similarity_dense_small(data.x)
        # print(sim.shape)
        for i in range(len(edge_dict)):
            neighbor_indices = np.array(edge_dict.get(i, []), dtype=np.int32)
            res_idx = torch.tensor(neighbor_indices)
            # sim_mean = 0
            if len(neighbor_indices) > 1:
                # continue
                idx = torch.tensor(neighbor_indices)
                new_sim = torch.index_select(torch.index_select(sim, 0, idx), 1, idx)
                new_sim[torch.eye(len(idx)).bool()] = 0.0
                # print(torch.sum(new_sim))
                sim_mean = torch.sum(new_sim) / (len(idx)*(len(idx)-1))

                if sim_mean > threshold:
                    for q, p in enumerate(res_idx):
                        pv_rows.append(n_count)
                        # pv_rows.append(i)
                        pv_cols.append(p)
                        if q == (len(res_idx)-1):
                            n_count += 1
            else:
                for q, p in enumerate(res_idx):
                    pv_rows.append(n_count)
                    # pv_rows.append(i)
                    pv_cols.append(p)
                    # if q == (len(res_idx)-1):
                    n_count += 1
        n_count2 = n_count
    elif -1 > threshold >= -2:

        n_count2 = n_count
        for i in range(len(edge_dict)):
            res_idx = np.array(edge_dict.get(i, []), dtype=np.int32)

            for q, p in enumerate(res_idx):
                pv_rows.append(n_count2)
                pv_cols.append(p)
                # if q == (len(res_idx)-1):
                n_count2 += 1
    else:
        n_count2 = n_count
        pv_init_val = []
        for i in range(len(edge_dict)):
            res_idx = np.array(edge_dict.get(i, []), dtype=np.int32)
            for q, p in enumerate(res_idx):
                if len(res_idx) == 1:
                    if res_idx[0] in node_num_egde:
                        if init_type == 1:
                            pv_init_val.append(node_num_egde[res_idx[0]]*init_val)
                        else:
                            pv_init_val.append(init_val)
                        # print()
                    else:
                        pv_init_val.append(1)
                else:
                    pv_init_val.append(1)
                pv_rows.append(n_count2)
                pv_cols.append(p)
                if q == (len(res_idx)-1):
                    n_count2 += 1
    pv_rows = torch.tensor(pv_rows)
    pv_cols = torch.tensor(pv_cols)
    pv_indices = torch.stack([pv_rows, pv_cols], dim=0)
    pv_values = torch.tensor(pv_init_val, dtype=torch.float32)

    Pv = torch.sparse_coo_tensor(pv_indices, pv_values, size=[n_count2, N_vertex])
    PvT = torch.sparse_coo_tensor(torch.stack([pv_cols, pv_rows], dim=0), pv_values, size=[N_vertex, n_count2])

    if norm_type == 0:
        Pv_col_sum = torch.sparse.sum(Pv, dim=1)
        Pv_diag_indices = Pv_col_sum.indices()[0]
        Pv_diag_values = Pv_col_sum.values()
        Pv_diag_values = torch.reciprocal(Pv_diag_values)
        Pv_diag = torch.sparse_coo_tensor(torch.stack([Pv_diag_indices, Pv_diag_indices]), Pv_diag_values,
                                        torch.Size([Pv_col_sum.shape[0], Pv_col_sum.shape[0]]))
        Pv_col_norm = torch.sparse.mm(Pv_diag, Pv)

        PvT_row_sum = torch.sparse.sum(PvT, dim=1)
        PvT_diag_indices = PvT_row_sum.indices()[0]
        PvT_diag_values = PvT_row_sum.values()
        PvT_diag_values = torch.reciprocal(PvT_diag_values)
        PvT_diag = torch.sparse_coo_tensor(torch.stack([PvT_diag_indices, PvT_diag_indices]), PvT_diag_values,
                                        torch.Size([PvT_row_sum.shape[0], PvT_row_sum.shape[0]]))
        # print(PvT.shape, PvT_diag.shape)
        PvT_row_norm = torch.sparse.mm(PvT_diag, PvT)
    elif norm_type == 1:
        Pv_col_sum = torch.sparse.sum(Pv, dim=0)
        Pv_diag_indices = Pv_col_sum.indices()[0]
        Pv_diag_values = Pv_col_sum.values()
        Pv_diag_values = torch.reciprocal(Pv_diag_values)
        Pv_diag = torch.sparse_coo_tensor(torch.stack([Pv_diag_indices, Pv_diag_indices]), Pv_diag_values,
                                            torch.Size([Pv_col_sum.shape[0], Pv_col_sum.shape[0]]))
        # print(Pv.shape, Pv_diag.shape)
        Pv_col_norm = torch.sparse.mm(Pv, Pv_diag)

        PvT_row_sum = torch.sparse.sum(PvT, dim=1)
        PvT_diag_indices = PvT_row_sum.indices()[0]
        PvT_diag_values = PvT_row_sum.values()
        PvT_diag_values = torch.reciprocal(PvT_diag_values)
        PvT_diag = torch.sparse_coo_tensor(torch.stack([PvT_diag_indices, PvT_diag_indices]), PvT_diag_values,
                                            torch.Size([PvT_row_sum.shape[0], PvT_row_sum.shape[0]]))
        # print(PvT.shape, PvT_diag.shape)
        PvT_row_norm = torch.sparse.mm(PvT_diag, PvT)
    elif norm_type == 2:
        Pv_col_sum = torch.sparse.sum(Pv, dim=0)
        Pv_diag_indices = Pv_col_sum.indices()[0]
        Pv_diag_values = Pv_col_sum.values()
        Pv_diag_values = torch.reciprocal(Pv_diag_values)
        Pv_diag = torch.sparse_coo_tensor(torch.stack([Pv_diag_indices, Pv_diag_indices]), Pv_diag_values,
                                            torch.Size([Pv_col_sum.shape[0], Pv_col_sum.shape[0]]))
        # print(Pv.shape, Pv_diag.shape)
        Pv_col_norm = torch.sparse.mm(Pv, Pv_diag)

        PvT_row_sum = torch.sparse.sum(PvT, dim=0)
        PvT_diag_indices = PvT_row_sum.indices()[0]
        PvT_diag_values = PvT_row_sum.values()
        PvT_diag_values = torch.reciprocal(PvT_diag_values)
        PvT_diag = torch.sparse_coo_tensor(torch.stack([PvT_diag_indices, PvT_diag_indices]), PvT_diag_values,
                                            torch.Size([PvT_row_sum.shape[0], PvT_row_sum.shape[0]]))
        # print(PvT.shape, PvT_diag.shape)
        PvT_row_norm = torch.sparse.mm(PvT, PvT_diag)
    elif norm_type == 3:
        Pv_col_sum = torch.sparse.sum(Pv, dim=1)
        Pv_diag_indices = Pv_col_sum.indices()[0]
        Pv_diag_values = Pv_col_sum.values()
        Pv_diag_values = torch.reciprocal(Pv_diag_values)
        Pv_diag = torch.sparse_coo_tensor(torch.stack([Pv_diag_indices, Pv_diag_indices]), Pv_diag_values,
                                            torch.Size([Pv_col_sum.shape[0], Pv_col_sum.shape[0]]))
        # print(Pv.shape, Pv_diag.shape)
        Pv_col_norm = torch.sparse.mm(Pv_diag, Pv)

        PvT_row_sum = torch.sparse.sum(PvT, dim=0)
        PvT_diag_indices = PvT_row_sum.indices()[0]
        PvT_diag_values = PvT_row_sum.values()
        PvT_diag_values = torch.reciprocal(PvT_diag_values)
        PvT_diag = torch.sparse_coo_tensor(torch.stack([PvT_diag_indices, PvT_diag_indices]), PvT_diag_values,
                                            torch.Size([PvT_row_sum.shape[0], PvT_row_sum.shape[0]]))

        PvT_row_norm = torch.sparse.mm(PvT, PvT_diag)
    else:
        Pv_col_norm = Pv
        PvT_row_norm = PvT

    print(Pv_col_norm.shape, data.x.shape)
    
    data.Pv,data.PvT = Pv_col_norm.to(args.device),PvT_row_norm.to(args.device)
    
    return data

#---------------------- LEGCN preprocessing utils -------------------------

def legcn_preprocessing(data,args):
    data.le_adj = line_expansion(data,args)
    return data

def line_expansion(data,args):
    hyperedge_index=copy.deepcopy(data.hyperedge_index)
    hyperedge_index = hyperedge_index.to('cpu')
    V, E = hyperedge_index
    E = E + V.max() # [V | E]
    num_ne_pairs = hyperedge_index.shape[1]
    V_plus_E = E.max() + 1
    L1 = torch.stack([torch.arange(V.shape[0], device=V.device), V], -1)
    L2 = torch.stack([torch.arange(E.shape[0], device=E.device), E], -1)
    L = torch.cat([L1, L2], -1) # [2, |V| + |E|]
    L_T = torch.stack([L[1], L[0]], 0) # [2, |V| + |E|]
    ones = torch.ones(L.shape[1], device=L.device)
    adj, value = torch_sparse.spspmm(L, ones, L_T, ones, num_ne_pairs, V_plus_E, num_ne_pairs, coalesced=True)
    adj, value = torch_sparse.coalesce(adj, value, num_ne_pairs, num_ne_pairs, op="add")
    return adj.to(args.device)

#---------------------- HJRL preprocessing utils -------------------------

def hjrl_preprocessing(data,args):
    
    hyperedge_index=copy.deepcopy(data.hyperedge_index)
    hyperedge_index = hyperedge_index.to('cpu')

    row = hyperedge_index[0].numpy()  
    col = hyperedge_index[1].numpy() 
    values = np.ones(len(row))  

    num_nodes = row.max() + 1 
    num_hyperedges = col.max() + 1  
    H_ini = sp.coo_matrix((values, (row, col)), shape=(num_nodes, num_hyperedges)).astype(np.float32) 

    HHT, H = normalize_sparse_hypergraph_symmetric(H_ini)
    HTH, HT = normalize_sparse_hypergraph_symmetric(H_ini.transpose())
    
    # 2. node/edge feature normalization
    X,E = precompute_node_edge_features(H_ini,data.x,args.device)
    
    HHT = ssm2tst(HHT).to(args.device)
    H = ssm2tst(H).to(args.device)
    HT = ssm2tst(HT).to(args.device)
    HTH = ssm2tst(HTH).to(args.device) 
    H_ini = torch.from_numpy(H_ini.toarray()).to(args.device)
    
    data.H_ini, data.H, data.HT, data.HHT, data.HTH, data.norm_X, data.norm_E = H_ini, H, HT, HHT, HTH, X, E

    return data

def normalize_sparse_hypergraph_symmetric(H):
    rowsum = np.array(H.sum(1))
    r_inv_sqrt = np.power(rowsum, -0.5).flatten()
    r_inv_sqrt[np.isinf(r_inv_sqrt)] = 0.
    D = sp.diags(r_inv_sqrt)

    colsum = np.array(H.sum(0),dtype='float')
    r_inv_sqrt = np.power(colsum, -1).flatten()
    r_inv_sqrt[np.isinf(r_inv_sqrt)] = 0.
    B = sp.diags(r_inv_sqrt)

    Omega = sp.eye(B.shape[0])

    hx1 = D.dot(H).dot(Omega).dot(B).dot(H.transpose()).dot(D)
    hx2 = D.dot(H).dot(Omega).dot(B)

    return hx1, hx2

def precompute_node_edge_features(H_ini,X,device):

    X = copy.deepcopy(X)
    X = X.cpu()
    E_node = H_ini.toarray().shape[1]
    H_T = H_ini.toarray().T
    E = np.zeros((E_node, X.shape[1]))
    for i in range(E_node):
        flag = np.where(H_T[i])[0]
        flag = X[flag]
        flag = flag.sum(0)
        E[i] = flag
    E = np.where(E, 1, 0).astype(float)
    #X = normalize_features(sp.csr_matrix(X))
    #E = normalize_features(sp.csr_matrix(E))
    X = row_normalize(sp.csr_matrix(X))
    E = row_normalize(sp.csr_matrix(E))
    
    X = torch.from_numpy(X.toarray()).float().to(device)
    E = torch.from_numpy(E.toarray()).float().to(device)

    return X,E

#---------------------- HyperGCN preprocessing utils -------------------------

# functions for processing/checkning the edge_index
def get_HyperGCN_He_dict_old(data):
    # Assume edge_index = [V;E], sorted
    edge_index = np.array(data.hyperedge_index.cpu())
    """
    For each he, clique-expansion. Note that we allow the weighted edge.
    Note that if node pair (vi,vj) is contained in both he1, he2, we will have (vi,vj) twice in edge_index. (weighted version CE)
    We default no self loops so far.
    """
    # edge_index[1, :] = edge_index[1, :]-edge_index[1, :].min()
    He_dict = {}
    for he in np.unique(edge_index[1, :]):
        nodes_in_he = list(edge_index[0, :][edge_index[1, :] == he])
        He_dict[he.item()] = nodes_in_he

    return He_dict

def get_HyperGCN_He_dict(data):
    edge_index = np.array(data.hyperedge_index.cpu())
    He_dict = defaultdict(list)

    for node, he in zip(edge_index[0, :], edge_index[1, :]):
        He_dict[he.item()].append(node)

    return dict(He_dict)

def update(Se, Ie, mediator, weights, c):
    """
    updates the weight on {Se,mediator} and {Ie,mediator}
    """    
    
    if (Se,mediator) not in weights:
        weights[(Se,mediator)] = 0
    weights[(Se,mediator)] += float(1/c)

    if (Ie,mediator) not in weights:
        weights[(Ie,mediator)] = 0
    weights[(Ie,mediator)] += float(1/c)

    if (mediator,Se) not in weights:
        weights[(mediator,Se)] = 0
    weights[(mediator,Se)] += float(1/c)

    if (mediator,Ie) not in weights:
        weights[(mediator,Ie)] = 0
    weights[(mediator,Ie)] += float(1/c)

    return weights

#---------------------- TMPHN preprocessing utils -------------------------

def tmphn_preprocessing(data,args):

    hyperedge_index=copy.deepcopy(data.hyperedge_index)
    hyperedge_index = hyperedge_index.to('cpu')

    row = hyperedge_index[0].numpy()  
    col = hyperedge_index[1].numpy()  
    values = np.ones(len(row))  

    num_nodes = row.max() + 1  
    num_hyperedges = col.max() + 1 
    H = sp.coo_matrix((values, (row, col)), shape=(num_nodes, num_hyperedges)).astype(np.float32) 
    H=H.toarray()
    
    all_nodes = list(np.arange(data.x.shape[0])) 
    neig_dict = NeighborFinder(H, args).neig_for_targets(all_nodes)
    data.neig_dict = neig_dict
    data.all_idx = np.arange(data.x.shape[0]) 
    data.target_idx = np.arange(data.x.shape[0]) 
    
    return data

class NeighborFinder:
    """
    Given a set of target nodes, find their neighbors
    args:
    edge_dict: an edge dictionary where key = edge_id, value = a list of nodes in this edge
    M: the maximum number of neighbors to be sampled for each target node (order of the hypergraph)
    return :
    a neighbhorhood dictinoary where key = a targe node, value = a nested list contains its neighbors
    """
    def __init__(self, H, args):
        self.H = H
        self.E = H.shape[1]
        self.M = args.M
        edge_dict = self.from_H_to_dict()
        self.adj_lst = [list(edge_dict[edge_id]) for edge_id in range(len(edge_dict))]#the serch space
        # self.adj_lst = adj_lst
        
    def neig_for_targets(self, target_nodes):
        """
        use dask to search over the adj_lst to find neighbors for all target nodes
        return: 
        batch_dict: a dictionary maps target_nodes to their neighbors
        """
        print('NeighborFinder Search: ')
        neig_list = []
        for x in target_nodes:
            y = delayed(self.find_neigs_of_one_node)(x)
            neig_list.append(y)
        with ProgressBar():
            neig_lst = dask.compute(neig_list, num_workers=os.cpu_count()*2)
        
        neig_lst = sum(neig_lst, []) #un-neste 
        
        batch_dict = dict(zip(target_nodes, neig_lst))

        return batch_dict
        
    def find_neigs_of_one_node(self, target_node):
        neigs_of_node = []
        for edge in self.adj_lst:
            if target_node in edge:
                if len(edge) <= self.M:
                    neigs_of_node.append(list(edge))
                else:
                    edge_lst = list(edge)
                    tmp = copy.deepcopy(edge_lst)
                    tmp.remove(target_node)
                    random.seed(42)
                    neigs_of_node.append(random.sample(tmp, self.M - 1) + [target_node])      
        return neigs_of_node
    
    def from_H_to_dict(self):
        """
        Take the incidence matrix as input, produce the incidence dictionary
        that will be used in message passing.
        Input: 
        H: incidence matrix, (N, E)
        Output:
        inci_dic: incidence dictionary with key = edge id, value = set(incident node idx)
        """
        edges_lst = [set(np.nonzero(self.H[:,i])[0]) for i in range(self.E)] #all edges
        edge_idx_lst = list(np.arange(0, self.E, 1))
        edge_dict = dict(map(lambda i,j : (i,j) , edge_idx_lst, edges_lst))
        return edge_dict
    
    def from_csc_H_to_dict(self):
        """
        Take the incidence matrix as input, produce the incidence dictionary
        that will be used in message passing.
        Input: 
        H: incidence matrix that is stored in csc format, (N, E)
        Output:
        inci_dic: incidence dictionary with key = edge id, value = set(incident node idx)
        
        Note: this function is used for the business datasets
        """
        edge_dict = {}
        for col in range(self.H.shape[1]): # go through each hyperedge
            nonzero_rows = list(self.H[:, col].indices) #get the nodes
            edge_dict[col] = nonzero_rows
        return edge_dict

#---------------------- EHNN preprocessing utils -------------------------

def ehnn_preprocessing(data,args,folder='lib_ehnn_cache'):
    
    src_data = copy.deepcopy(data)
    data = data.to('cpu')
    
    # build memory-efficiently for yelp and walmart-trips-100
    if args.dname not in ["yelp", "walmart-trips-100"]:
        original_edge_index = data.hyperedge_index
        data = ConstructH(data)  # [|V|, |E|]
        incidence_d = (
            torch.tensor(data.hyperedge_index, dtype=torch.float32)
            .to_sparse(2)
            .coalesce()
            .to(args.device)
        )
        edge_orders_d = (
            torch.sparse.sum(incidence_d, 0).to_dense().long().to(args.device)
        )  # [|E|,]

        data.hyperedge_index = original_edge_index
        data = ConstructHSparse(data)
        num_nodes = data.x.shape[0]
        num_hyperedges = np.max(data.hyperedge_index[1]) + 1
        incidence_s = torch.sparse_coo_tensor(
            data.hyperedge_index,
            torch.ones(len(data.hyperedge_index[0])),
            (num_nodes, num_hyperedges),
            device=args.device,
        ).coalesce()
        edge_orders_s = (
            torch.sparse.sum(incidence_s, 0).to_dense().long().to(args.device)
        )  # [|E|,]

        assert (incidence_d.indices() - incidence_s.indices() == 0).all()
        assert (incidence_d.values() - incidence_s.values() == 0).all()

        incidence = incidence_d
        edge_orders = edge_orders_d
    else:
        data = ConstructHSparse(data)
        num_nodes = data.x.shape[0]
        num_hyperedges = np.max(data.hyperedge_index[1]) + 1
        incidence = (
            torch.sparse_coo_tensor(
                data.hyperedge_index,
                torch.ones(len(data.hyperedge_index[0])),
                (num_nodes, num_hyperedges),
            )
            .coalesce()
            .to(args.device)
        )
        edge_orders = torch.sparse.sum(incidence, 0).to_dense().long().to(args.device)

    os.makedirs(f"./{folder}", exist_ok=True)
    
    if not osp.isfile(f"./{folder}/{args.dname}.pt") or args.task_type=='hg_cls':
        print(f"preprocessing {args.dname}")
        overlaps = None
        masks = None
        n_overlaps = None
        prefix_normalizer = (torch.sparse.sum(incidence, 0).to_dense().to(args.device))  # [|E|,]
        prefix_normalizer = prefix_normalizer.masked_fill_(prefix_normalizer == 0, 1e-5)
        normalizer = None
        suffix_normalizer = (torch.sparse.sum(incidence, 1).to_dense().to(args.device))  # [|V|,]
        suffix_normalizer = suffix_normalizer.masked_fill_(suffix_normalizer == 0, 1e-5)

        if args.dname not in ["yelp", "walmart-trips-100"]:
            # chunked mask computation
            mask_dict_chunk = build_mask_chunk(incidence, args.device, args.chunk_size)  # Dict(overlap: [|E|, |E|] sparse)
            overlaps_chunk, masks_chunk = list(mask_dict_chunk.keys()), list(mask_dict_chunk.values())
            overlaps_chunk = torch.tensor(overlaps_chunk, dtype=torch.long, device=args.device)  # [|overlaps|,]
            masks_chunk = torch.stack(masks_chunk, dim=0).coalesce()  # [|overlaps|, |E|, |E|]]

            # correctness check with non-chunked masks
            mask_dict = build_mask(incidence, args.device)  # Dict(overlap: [|E|, |E|] sparse)
            overlaps, masks = list(mask_dict.keys()), list(mask_dict.values())
            overlaps = torch.tensor(overlaps, dtype=torch.long, device=args.device)  # [|overlaps|,]
            masks = torch.stack(masks, dim=0).coalesce()  # [|overlaps|, |E|, |E|]]
            assert (masks.indices() - masks_chunk.indices() == 0).all()
            assert (masks.values() - masks_chunk.values() == 0).all()
            assert (overlaps - overlaps_chunk == 0).all()

            masks = masks_chunk
            overlaps = overlaps_chunk
            n_overlaps = len(overlaps)

            normalizer = (torch.sparse.sum(masks, 2).to_dense().unsqueeze(-1))  # [|overlaps|, |E|, 1]
            normalizer = normalizer.masked_fill_(normalizer == 0, 1e-5)

        ehnn_cache = dict(
            incidence=incidence,
            edge_orders=edge_orders,
            overlaps=overlaps,
            n_overlaps=n_overlaps,
            prefix_normalizer=prefix_normalizer,
            suffix_normalizer=suffix_normalizer,
        )

        torch.save(ehnn_cache, f"./{folder}/{args.dname}.pt")
        print(f"saved ehnn_cache for {args.dname}")
    else:
        ehnn_cache = torch.load(f"./{folder}/{args.dname}.pt")
    print(f"number of mask channels: {ehnn_cache['n_overlaps']}")
    
    src_data.ehnn_cache = ehnn_cache
    
    return src_data

@torch.no_grad()
def build_mask_chunk(incidence, device, chunk_size = 1000):
    tic = time.time()
    stat_overlap_num = {}  # key: overlap, value: number of edges
    incidence = incidence.to(device)
    N = incidence.size(0)
    E = incidence.size(1)
    chunk_begin_index = 0
    output = torch.sparse_coo_tensor(
        torch.empty([2, 0]), [], [E, E], device=device
    ).coalesce()
    target_indices = incidence.indices()[
        [1, 0]
    ]  # caution: indices of transposed incidence matrix
    row_indices = target_indices[0]  # [|E|,]
    target_values = incidence.values()  # [|E|,]

    while chunk_begin_index < E:
        chunk_end_index = min(chunk_begin_index + chunk_size, incidence.size(1))
        chunk_mask = torch.logical_and(
            row_indices >= chunk_begin_index, row_indices < chunk_end_index
        )  # [|chunk|]
        chunk_matrix = torch.sparse_coo_tensor(
            target_indices[:, chunk_mask],
            target_values[chunk_mask],
            [E, N],
            device=device,
        ).coalesce()  # [2, |chunk|]
        del chunk_mask
        print(
            f"chunk_begin_index: {chunk_begin_index}/{E}, "
            f"output: {output.indices().size(1)} nnz, "
            f"chunk_matrix: {chunk_matrix.indices().size(1)} nnz"
        )
        chunk_output = torch.sparse.mm(chunk_matrix, incidence)
        output = (output + chunk_output).coalesce()
        del chunk_matrix
        chunk_begin_index += chunk_size

    overlap_mask_indices = defaultdict(lambda: [[], []])
    for overlap, row_idx, col_idx in zip(
        output.values(), output.indices()[0], output.indices()[1]
    ):
        overlap = int(overlap.item())
        overlap_mask_indices[overlap][0].append(row_idx.item())
        overlap_mask_indices[overlap][1].append(col_idx.item())

    overlap_masks = {}
    for overlap, mask_indices in overlap_mask_indices.items():
        overlap_masks[overlap] = torch.sparse_coo_tensor(
            mask_indices, torch.ones(len(mask_indices[0])), (E, E), device=device
        )
    print(f"Took {time.time() - tic} seconds")
    return overlap_masks

@torch.no_grad()
def build_mask(incidence, device):
    tic = time.time()
    overlap_masks = {}
    # stat_overlap_num = {}  # Dict(overlap: number of edges)

    incidence = incidence.to(device)

    E = incidence.size(1)
    pre_overlap = torch.sparse.mm(incidence.t(), incidence).coalesce()  # [|E|, |E|]
    # stat_num_entry = len(pre_overlap.indices()[0])
    overlap_mask_indices = defaultdict(lambda: [[], []])

    # Generate Overlap masks - Not made for overlap mask of overlap 0
    for overlap, row_idx, col_idx in zip(
        pre_overlap.values(), pre_overlap.indices()[0], pre_overlap.indices()[1]
    ):
        overlap = int(overlap.item())
        overlap_mask_indices[overlap][0].append(row_idx.item())
        overlap_mask_indices[overlap][1].append(col_idx.item())

    for overlap, mask_indices in overlap_mask_indices.items():
        overlap_masks[overlap] = torch.sparse_coo_tensor(
            mask_indices, torch.ones(len(mask_indices[0])), (E, E), device=device
        )
    print(f"Took {time.time() - tic} seconds")

    return overlap_masks

def ConstructH(data):
    """
    Construct incidence matrix H of size (num_nodes,num_hyperedges) from edge_index = [V;E]
    """
    edge_index = np.array(data.hyperedge_index)
    num_nodes = data.x.shape[0]
    num_hyperedges = np.max(edge_index[1]) - np.min(edge_index[1]) + 1
    H = np.zeros((num_nodes, num_hyperedges))
    cur_idx = 0
    for he in np.unique(edge_index[1]):
        nodes_in_he = edge_index[0][edge_index[1] == he]
        H[nodes_in_he, cur_idx] = 1.0
        cur_idx += 1

    data.hyperedge_index = H
    return data

def ConstructHSparse(data):
    """
    Construct incidence matrix H of size (num_nodes,num_hyperedges) from edge_index = [V;E]
    """
    edge_index = np.array(data.hyperedge_index)
    num_nodes = data.x.shape[0]
    num_hyperedges = np.max(edge_index[1]) - np.min(edge_index[1]) + 1
    assert np.min(edge_index[1]) == 0 and np.max(edge_index[1]) == num_hyperedges - 1
    data.hyperedge_index = edge_index
    return data

#---------------------- HyperGT preprocessing utils -------------------------

def hypergt_preprocessing(data, args, add_reverse=False, add_token_loops=True):
    """
      output: 
      - data.num_tokens = n + m
      - data.H:    [n, m] (float32, device)
      - data.adjs: [edge_index_token] (long, device)
      - data.x:    [n+m, d]
    """
    dev = torch.device(args.device)

    ei = data.hyperedge_index.to(torch.long).to(dev)   # [2, E]
    v  = ei[0]                                         # node id
    e0 = ei[1]                                         # hyperedge id
    n  = int(getattr(data, 'num_nodes', data.x.size(0)))

    e_min = int(e0.min().item())
    e     = e0 - e_min                                 # 0..m-1
    m     = int(e.max().item()) + 1
    N     = n + m

    # ---- 2) construct H: [n, m] (float32, device) ----
    vals = torch.ones(e.size(0), dtype=torch.float32, device=dev)
    H = torch.sparse_coo_tensor(
        indices=torch.stack([v, e], dim=0),            # [2, E]
        values=vals,
        size=(n, m),
        dtype=torch.float32,
        device=dev
    ).coalesce()

    # ---- 3) construct token adjs[0]（V->[n..n+m-1]）----
    e_off = e + n                                   
    edge_v2e = torch.stack([v, e_off], dim=0)          # [2, E]
    edge_v2e, _ = remove_self_loops(edge_v2e)

    if add_reverse:
        edge_e2v = torch.stack([e_off, v], dim=0)      # [2, E]
        edge_index = torch.cat([edge_v2e, edge_e2v], dim=1)  # [2, 2E]
    else:
        edge_index = edge_v2e

    if add_token_loops:
        loops = torch.arange(N, device=dev, dtype=torch.long)
        loop_edge = torch.stack([loops, loops], dim=0)       # [2, N]
        edge_index = torch.cat([edge_index, loop_edge], dim=1)

    edge_index = edge_index.to(torch.long).to(dev)

    # ---- 4) x -> [n, m]
    x = data.x
    if x.device != dev:
        x = x.to(dev)
    if x.size(0) == n:
        pad = torch.zeros((m, x.size(1)), dtype=x.dtype, device=dev)
        x = torch.cat([x, pad], dim=0)                       # [n+m, F]
    elif x.size(0) != N:
        raise ValueError(f"data.x.shape[0]={x.size(0)} not match n(={n})/m(={m})")

    data.num_tokens = N
    data.H = H
    data.adjs = [edge_index]
    data.x = x

    return data


import torch
import torch.nn as nn
from dhgbench.lib_models.HNN.mlp import MLP
import numpy as np
import scipy.sparse as sp
from dhgbench.lib_models.HNN.utils import create_coo_from_edge_index,ssm2tst

class TFHNN(nn.Module):
    
    def __init__(self, num_features, num_targets, args):

        super().__init__()

        self.L = args.All_num_layers
        self.num_decoder_layers = args.MLP_num_layers 
        self.hidden_dim = args.MLP_hidden
        self.alpha = args.alpha
        self.dropout = args.dropout
        self.norm = args.normalization
        self.args = args

        self.cache = None

        self.decoder = MLP(in_channels=num_features,
            hidden_channels=self.hidden_dim,
            out_channels=num_targets,
            num_layers=self.num_decoder_layers,
            dropout=self.dropout,
            Normalization=self.norm,
            InputNorm=False)

    def reset_parameters(self):
        self.cache = None
        self.decoder.reset_parameters()

    def weighted_ce_graph(self,data,args):

        H = create_coo_from_edge_index(data.hyperedge_index) 

        # 1. H x B x H^{T}
        colsum = np.array(H.sum(0),dtype='float')
        r_inv_sqrt = np.power(colsum, -1).flatten()
        r_inv_sqrt[np.isinf(r_inv_sqrt)] = 0.
        B = sp.diags(r_inv_sqrt)

        A_beta = H.dot(B).dot(H.transpose())
        I = sp.eye(A_beta.shape[0])
        A_beta += I 

        # 2. D^{-1/2} x H x B x H^{T} x D^{1/2}
        rowsum = np.array(A_beta.sum(1),dtype='float')
        r_inv_sqrt = np.power(rowsum, -0.5).flatten()
        r_inv_sqrt[np.isinf(r_inv_sqrt)] = 0.
        D = sp.diags(r_inv_sqrt)

        A_beta = D.dot(A_beta).dot(D)
        
        return ssm2tst(A_beta).to(args.device)
    
    def diffusion(self, data):
        
        if self.cache is not None:
            return self.cache
            
        X = data.x
        
        norm_W = self.weighted_ce_graph(data,self.args)
        
        assert norm_W.layout == torch.sparse_coo
        norm_W = norm_W.coalesce()

        N = norm_W.size(0)
        X_out = self.alpha * X  # l = 0: (1 - alpha)^0 * X

        W_power_X = X.clone()

        for l in range(1, self.L):
            W_power_X = torch.sparse.mm(norm_W, W_power_X)
            X_out += self.alpha * ((1 - self.alpha) ** l) * W_power_X

        W_L_X = torch.sparse.mm(norm_W, W_power_X)
        X_out += ((1 - self.alpha) ** self.L) * W_L_X

        self.cache = X_out

        return X_out

    def forward(self, data):
        F = self.diffusion(data)
        return self.decoder(F), None 